# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main module for PyTree checkpoint metadata storage."""

from __future__ import annotations

import asyncio
import collections
import dataclasses
import enum
import functools
import inspect
import json
import operator
import typing
from typing import Any, Dict, Hashable, List, Optional, Protocol, Tuple, TypeAlias, TypeVar, Union

from absl import logging
from etils import epath
import jax
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import pytree_metadata_options as pytree_metadata_options_lib
from orbax.checkpoint._src.metadata import tree_rich_types
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.metadata import value_metadata_entry
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
from orbax.checkpoint._src.serialization import types
from orbax.checkpoint._src.tree import types as tree_types
from orbax.checkpoint._src.tree import utils as tree_utils


ValueMetadataEntry: TypeAlias = value_metadata_entry.ValueMetadataEntry
PyTreeMetadataOptions: TypeAlias = (
    pytree_metadata_options_lib.PyTreeMetadataOptions
)
PyTree: TypeAlias = Any
KeyEntry = TypeVar('KeyEntry', bound=Hashable)
KeyPath: TypeAlias = tuple[KeyEntry, ...]

PYTREE_METADATA_OPTIONS = pytree_metadata_options_lib.PYTREE_METADATA_OPTIONS

_KEY_NAME = 'key'
_KEY_TYPE = 'key_type'
_TREE_METADATA_KEY = 'tree_metadata'
_KEY_METADATA_KEY = 'key_metadata'
_VALUE_METADATA_KEY = 'value_metadata'
_USE_ZARR3 = 'use_zarr3'
_STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE = 'store_array_data_equal_to_fill_value'
_VALUE_METADATA_TREE = 'value_metadata_tree'
_CUSTOM_METADATA = 'custom_metadata'


class KeyType(enum.Enum):
  """Enum representing PyTree key type."""

  SEQUENCE = 1
  DICT = 2

  def to_json(self) -> int:
    return self.value

  @classmethod
  def from_json(cls, value: int) -> KeyType:
    return cls(value)


def _get_key_metadata_type(key: Any) -> KeyType:
  """Translates the JAX key class into a proto enum."""
  if tree_utils.is_sequence_key(key):
    return KeyType.SEQUENCE
  elif tree_utils.is_dict_key(key):
    return KeyType.DICT
  else:
    raise ValueError(f'Unsupported KeyEntry: {type(key)}: "{key}"')


def _keypath_from_key_type(key_name: str, key_type: KeyType) -> Any:
  """Converts from Key in InternalTreeMetadata to JAX keypath class."""
  if key_type == KeyType.SEQUENCE:
    return jax.tree_util.SequenceKey(int(key_name))
  elif key_type == KeyType.DICT:
    return jax.tree_util.DictKey(key_name)
  else:
    raise ValueError(f'Unsupported KeyEntry: {key_type}')


@dataclasses.dataclass
class NestedKeyMetadataEntry:
  """Represents a key at a single level of nesting."""

  nested_key_name: str
  key_type: KeyType

  def to_json(self) -> Dict[str, Union[str, int]]:
    return {
        _KEY_NAME: self.nested_key_name,
        _KEY_TYPE: self.key_type.to_json(),
    }

  @classmethod
  def from_json(
      cls, json_dict: Dict[str, Union[str, int]]
  ) -> NestedKeyMetadataEntry:
    return NestedKeyMetadataEntry(
        nested_key_name=json_dict[_KEY_NAME],
        key_type=KeyType.from_json(json_dict[_KEY_TYPE]),
    )


@dataclasses.dataclass
class KeyMetadataEntry:
  """Represents metadata for a key (all levels of nesting)."""

  nested_key_metadata_entries: List[NestedKeyMetadataEntry]

  def to_json(self) -> Tuple[Dict[str, Union[str, int]], ...]:
    return tuple(
        [entry.to_json() for entry in self.nested_key_metadata_entries]
    )

  @classmethod
  def from_json(
      cls, json_dict: Tuple[Dict[str, Union[str, int]], ...]
  ) -> KeyMetadataEntry:
    return KeyMetadataEntry(
        [NestedKeyMetadataEntry.from_json(entry) for entry in json_dict]
    )

  @classmethod
  def build(cls, keypath: KeyPath) -> KeyMetadataEntry:
    return KeyMetadataEntry([
        NestedKeyMetadataEntry(
            str(tree_utils.get_key_name(k)), _get_key_metadata_type(k)
        )
        for k in keypath
    ])


@dataclasses.dataclass
class InternalTreeMetadataEntry:
  """Represents metadata for a named key/value pair in a tree."""

  keypath: str
  key_metadata: KeyMetadataEntry
  value_metadata: ValueMetadataEntry

  def to_json(self) -> Dict[str, Any]:
    return {
        self.keypath: {
            _KEY_METADATA_KEY: self.key_metadata.to_json(),
            _VALUE_METADATA_KEY: self.value_metadata.to_json(),
        }
    }

  @classmethod
  def from_json(
      cls, keypath: str, json_dict: Dict[str, Any]
  ) -> InternalTreeMetadataEntry:
    return InternalTreeMetadataEntry(
        keypath,
        KeyMetadataEntry.from_json(json_dict[_KEY_METADATA_KEY]),
        ValueMetadataEntry.from_json(
            json_dict[_VALUE_METADATA_KEY],
            pytree_metadata_options=PyTreeMetadataOptions(
                support_rich_types=False  # Always in legacy mode.
            ),
        ),
    )

  @classmethod
  def build(
      cls,
      keypath: KeyPath,
      info: types.ParamInfo,
      save_arg: types.SaveArgs,
  ) -> InternalTreeMetadataEntry:
    """Builds a InternalTreeMetadataEntry."""
    return InternalTreeMetadataEntry(
        keypath=str(tuple([str(tree_utils.get_key_name(k)) for k in keypath])),
        key_metadata=KeyMetadataEntry.build(keypath),
        value_metadata=ValueMetadataEntry.build(info, save_arg),
    )

  def jax_keypath(self) -> KeyPath:
    keypath = []
    for nested_key_entry in self.key_metadata.nested_key_metadata_entries:
      nested_key_name = nested_key_entry.nested_key_name
      key_type = nested_key_entry.key_type
      keypath.append(_keypath_from_key_type(nested_key_name, key_type))
    return tuple(keypath)


@dataclasses.dataclass(kw_only=True)
class InternalTreeMetadata:
  """Metadata representation of a PyTree.

  Corresponds to the metadata of a PyTree checkpoint (e.g. that saved by
  `PyTreeCheckpointHandler`, `StandardCheckpointHandler`,
  `StandardCheckpointer`, etc.).

  This class is the internal / on-disk representation of metadata that is
  presented to the user as a `TreeMetadata` object.
  """

  tree_metadata_entries: List[InternalTreeMetadataEntry]
  use_zarr3: bool
  custom_metadata: tree_types.JsonType | None
  store_array_data_equal_to_fill_value: bool
  pytree_metadata_options: PyTreeMetadataOptions
  value_metadata_tree: PyTree | None = None

  def __post_init__(self):
    logging.vlog(
        1,
        'Created InternalTreeMetadata with pytree_metadata_options=%s, has'
        ' tree_metadata_entries=%s, has rich typed value_metadata_tree=%s',
        self.pytree_metadata_options,
        len(self.tree_metadata_entries),
        self.value_metadata_tree is not None,
    )
    # Validate JSON-serializability of custom_metadata.
    try:
      json.dumps(self.custom_metadata)
    except TypeError as e:
      raise TypeError(
          'Failed to encode `custom_metadata` metadata as JSON object. Please'
          ' ensure your `custom_metadata` is JSON-serializable.'
      ) from e

  @classmethod
  def build(
      cls,
      param_infos: PyTree,
      *,
      save_args: Optional[PyTree] = None,
      use_zarr3: bool = False,
      custom_metadata: tree_types.JsonType | None = None,
      pytree_metadata_options: PyTreeMetadataOptions = (
          PYTREE_METADATA_OPTIONS
      ),
  ) -> InternalTreeMetadata:
    """Returns an InternalTreeMetadata instance."""
    if save_args is None:
      save_args = jax.tree.map(
          lambda _: types.SaveArgs(),
          param_infos,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
    flat_info_with_keys, _ = jax.tree_util.tree_flatten_with_path(
        param_infos, is_leaf=tree_utils.is_empty_or_leaf
    )
    flat_save_args_with_keys, _ = jax.tree_util.tree_flatten_with_path(
        save_args, is_leaf=tree_utils.is_empty_or_leaf
    )
    tree_metadata_entries = []
    for (keypath, info), (_, save_arg) in zip(
        flat_info_with_keys, flat_save_args_with_keys
    ):
      tree_metadata_entries.append(
          InternalTreeMetadataEntry.build(keypath, info, save_arg)
      )
    value_metadata_tree = None
    if pytree_metadata_options.support_rich_types:
      value_metadata_tree = jax.tree_util.tree_map(
          ValueMetadataEntry.build,
          param_infos,
          save_args,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
      logging.vlog(
          1,
          'Created rich typed value_metadata_tree from param_infos and'
          ' save_args.',
      )
    return InternalTreeMetadata(
        tree_metadata_entries=tree_metadata_entries,
        use_zarr3=use_zarr3,
        custom_metadata=custom_metadata,
        store_array_data_equal_to_fill_value=ts_utils.STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE,
        pytree_metadata_options=pytree_metadata_options,
        value_metadata_tree=value_metadata_tree,
    )

  def to_json(self) -> Dict[str, Any]:
    """Returns a JSON representation of the metadata.

    Uses JSON format::
    ```
      {
          _TREE_METADATA_KEY: {
            "(top_level_key, lower_level_key)": {
                _KEY_METADATA_KEY: (
                    {_KEY_NAME: "top_level_key", _KEY_TYPE: <KeyType
                    (int)>},
                    {_KEY_NAME: "lower_level_key", _KEY_TYPE: <KeyType
                    (int)>},
                )
                _VALUE_METADATA_KEY: {
                    _VALUE_TYPE: "jax.Array",
                    _SKIP_DESERIALIZE: True/False,
                }
            }
            ...
          },
          _USE_ZARR3: True/False,
          _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE: True,
          _CUSTOM_METADATA: ...,
          _VALUE_METADATA_TREE: '{
            "mu_nu": {
              "category": "namedtuple",
              "module": "orbax.checkpoint._src.testing.test_tree_utils",
              "clazz": "MuNu",
              "entries": [
                {
                  "key": "mu",
                  "value": {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "jax.Array",
                      "skip_deserialize": false
                    }
                  }
                },
                {
                  "key": "nu",
                  "value": {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "np.ndarray",
                      "skip_deserialize": false
                    }
                  }
                }
              ]
            },
            "my_tuple": {
              "category": "custom",
              "clazz": "tuple",
              "entries": [
                  {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "np.ndarray",
                      "skip_deserialize": false
                    }
                  }
              ]
            }
          }'
      }
    ```
    """
    json_object = {
        _TREE_METADATA_KEY: functools.reduce(
            operator.ior,
            [entry.to_json() for entry in self.tree_metadata_entries],
            {},
        ),
        _USE_ZARR3: self.use_zarr3,
        _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE: (
            self.store_array_data_equal_to_fill_value
        ),
        _CUSTOM_METADATA: self.custom_metadata,
    }
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    if (
        self.pytree_metadata_options.support_rich_types
        and self.value_metadata_tree is not None
    ):
      json_object[_VALUE_METADATA_TREE] = (
          tree_rich_types.value_metadata_tree_to_json_str(
              self.value_metadata_tree
          )
      )
      logging.vlog(
          1, 'Serialized rich typed value_metadata_tree to json_object.'
      )
    return json_object

  @classmethod
  def from_json(
      cls,
      json_dict: Dict[str, Any],
      pytree_metadata_options: PyTreeMetadataOptions = (
          PYTREE_METADATA_OPTIONS
      ),
  ) -> InternalTreeMetadata:
    """Returns an InternalTreeMetadata instance from its JSON representation."""
    use_zarr3 = json_dict.get(_USE_ZARR3, False)
    custom_metadata = json_dict.get(_CUSTOM_METADATA, None)
    store_array_data_equal_to_fill_value = json_dict.get(
        _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE, False
    )

    tree_metadata_entries = []
    for keypath, json_tree_metadata_entry in json_dict[
        _TREE_METADATA_KEY
    ].items():
      tree_metadata_entries.append(
          InternalTreeMetadataEntry.from_json(keypath, json_tree_metadata_entry)
      )
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    value_metadata_tree = None
    if (
        pytree_metadata_options.support_rich_types
        and _VALUE_METADATA_TREE in json_dict
    ):
      value_metadata_tree = tree_rich_types.value_metadata_tree_from_json_str(
          json_dict[_VALUE_METADATA_TREE]
      )
      logging.info(
          'Deserialized rich typed value_metadata_tree from json_dict.'
      )
    return InternalTreeMetadata(
        tree_metadata_entries=tree_metadata_entries,
        use_zarr3=use_zarr3,
        custom_metadata=custom_metadata,
        pytree_metadata_options=pytree_metadata_options,
        value_metadata_tree=value_metadata_tree,
        store_array_data_equal_to_fill_value=store_array_data_equal_to_fill_value,
    )

  def as_nested_tree(self) -> Dict[str, Any]:
    """Converts to a nested tree, with leaves of ValueMetadataEntry."""
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    if (
        self.pytree_metadata_options.support_rich_types
        and self.value_metadata_tree is not None
    ):
      logging.info(
          'Returning rich typed value_metadata_tree from InternalTreeMetadata.'
      )
      return self.value_metadata_tree

    return tree_utils.from_flattened_with_keypath([
        (entry.jax_keypath(), entry.value_metadata)
        for entry in self.tree_metadata_entries
    ])

  def as_custom_metadata(
      self,
      directory: epath.Path,
      type_handler_registry: types.TypeHandlerRegistry,
      *,
      use_ocdbt: bool = True,
  ) -> PyTree:
    """Returns a user-facing PyTree with leaves of `ValueMetadataEntry`.

    The returned PyTree conforms to the structure of `self` InternalTreeMetadata
    but the `ValueMetadataEntry` leaves are derived from the checkpoints in
    `directory`.

    Args:
      directory: The directory to read the checkpoint from.
      type_handler_registry: `TypeHandlerRegistry` whose registered
        TypeHandlers' metadata are used to build the `ValueMetadataEntry`
        leaves. See `TypeHandler.metadata()`.
      use_ocdbt: Whether to use OCDBT for reading the metadata from tensorstore.
    """
    flat_param_infos = {}
    flat_restore_types = {}
    reference_metadata_tree = self.as_nested_tree()
    ts_context = ts_utils.get_ts_context(use_ocdbt=use_ocdbt)
    for keypath, value_meta in tree_utils.to_flat_dict(
        reference_metadata_tree
    ).items():
      param_name = '.'.join(keypath)
      flat_param_infos[keypath] = types.ParamInfo(
          name=param_name,
          path=directory / param_name,
          parent_dir=directory,
          skip_deserialize=value_meta.skip_deserialize,
          is_ocdbt_checkpoint=use_ocdbt,
          use_zarr3=self.use_zarr3,
          ts_context=ts_context,
          write_shape=value_meta.write_shape,
      )
      flat_restore_types[keypath] = value_meta.value_type

    flat_metadatas = {}
    batched_param_infos = collections.defaultdict(list)
    batched_keypaths = collections.defaultdict(list)
    for keypath in flat_param_infos:
      param_info = flat_param_infos[keypath]
      restore_type = flat_restore_types[keypath]
      if param_info.skip_deserialize:
        if empty_values.is_empty_typestr(restore_type):
          flat_metadatas[keypath] = empty_values.get_empty_value_from_typestr(
              restore_type, self.pytree_metadata_options
          )
        else:
          flat_metadatas[keypath] = value_metadata.Metadata(
              name=param_info.name, directory=param_info.parent_dir
          )
      else:
        batched_keypaths[restore_type].append(keypath)
        batched_param_infos[restore_type].append(param_info)

    metadata_ops = []
    for restore_type, param_infos in batched_param_infos.items():
      handler = type_handler_registry.get(restore_type)
      metadata_ops.append(handler.metadata(param_infos))

    async def _get_metadata():
      return await asyncio.gather(*metadata_ops)

    batched_metadatas = asyncio_utils.run_sync(_get_metadata())
    for keypath_batch, metadata_batch in zip(
        batched_keypaths.values(), batched_metadatas
    ):
      for keypath, value in zip(keypath_batch, metadata_batch):
        flat_metadatas[keypath] = value
    return tree_utils.from_flat_dict(
        flat_metadatas, target=reference_metadata_tree
    )


def serialize_tree(
    tree: PyTree,
    pytree_metadata_options: PyTreeMetadataOptions,
) -> PyTree:
  """Transforms a PyTree to a serializable format.

  IMPORTANT: If `pytree_metadata_options.support_rich_types` is false, the
  returned tree replaces tuple container nodes with list nodes.

  IMPORTANT: If `pytree_metadata_options.support_rich_types` is false, the
  returned tree replaces NamedTuple container nodes with dict
  nodes.

  If `pytree_metadata_options.support_rich_types` is true, then the returned
  tree is the same as the input tree retaining empty nodes as leafs.

  Args:
    tree: The tree to serialize.
    pytree_metadata_options: `PyTreeMetadataOptions` for managing PyTree
      metadata.

  Returns:
    The serialized PyTree.
  """
  if pytree_metadata_options.support_rich_types:
    return jax.tree_util.tree_map(
        lambda x: x,
        tree,
        is_leaf=tree_utils.is_empty_or_leaf,
    )

  return tree_utils.serialize_tree(tree, keep_empty_nodes=True)


@typing.runtime_checkable
class TreeMetadata(Protocol):
  """User-facing metadata representation of a PyTree.

  Corresponds to the metadata of a PyTree checkpoint (e.g. that saved by
  `PyTreeCheckpointHandler`, `StandardCheckpointHandler`,
  `StandardCheckpointer`, etc.).

  Implementations must register themselves as PyTrees, e.g. with
  `jax.tree_util.register_pytree_with_keys_class`.

  The object should be treated as a regular PyTree that can be mapped over.
  Leaf values are typically of type `ocp.metadata.value.Metadata` (when the
  object is obtained from a `metadata()` function call). Note that the user may
  subsequently modify these leaves to be of any type. Additional properties
  (e.g. `custom_metadata`) may be accessed directly, and are independent from
  the
  tree structure. To directly access the underlying PyTree, which matches the
  checkpoint structure, use the `tree` property.

  Here is a typical example usage::
    with ocp.StandardCheckpointer() as ckptr:
      # `metadata` is a `TreeMetadata` object, but can be treated as a regular
      # PyTree. In this case, it corresponds to a "serialized" representation of
      # the checkpoint tree. This means that all custom_metadata nodes are
      converted to
      # standardized containers like list, tuple, and dict. (See also
      # `support_rich_types` for further details on how other types are
      # handled.)
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      # Access array properties.
      metadata['step'].shape
      metadata['step'].dtype
      # Access a list element.
      metadata['opt_state'][0]
      # Get all the shapes of the tree elements.
      shapes = jax.tree.map(lambda x: x.shape, metadata)

  If the checkpoint structure is standardized as a list or a tuple, the metadata
  object can be indexed like a regular sequence::
    with ocp.StandardCheckpointer() as ckptr:
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      metadata[0].shape
      shapes = jax.tree.map(lambda x: x.shape, metadata)

  Note that if we manually construct a target tree with the same structure as
  the checkpoint, we will run into an error if we try to tree map over it at the
  same time as the metadata object. To do this, instead access the `tree`
  property.

  For example, if our checkpoint has a structure like the following::

    {
        'a': 1,
        'b': 2,
    }

  The following will raise a tree mismatch error::

    with ocp.StandardCheckpointer() as ckptr:
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      tree = {
          'a': 1,
          'b': 2,
      }
      jax.tree.map(lambda x, y: foo(x, y), metadata, tree)

  To avoid this, use the `tree` property::

    with ocp.StandardCheckpointer() as ckptr:
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      tree = {
          'a': 1,
          'b': 2,
      }
      jax.tree.map(lambda x, y: foo(x, y), metadata.tree, tree)


  Properties of the `TreeMetadata` object, such as `custom_metadata` and `tree`,
  can be
  accessed directly::
    with ocp.StandardCheckpointer() as ckptr:
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      metadata.custom_metadata
      metadata.tree

  The metadata can be used directly to restore a checkpoint. Restoration code
  automatically extracts the necessary properties from the metadata. This is
  allowed because the `TreeMetadata` is by definition a PyTree matching the
  structure of the checkpoint::

    with ocp.StandardCheckpointer() as ckptr:
      metadata = ckptr.metadata('/path/to/existing/checkpoint')
      restored = ckptr.restore(
          '/path/to/existing/checkpoint',
          metadata,
      )
  """

  @property
  def tree(self) -> PyTree:
    ...

  @property
  def custom_metadata(self) -> PyTree | None:
    ...


  def tree_flatten(self):
    ...

  def tree_flatten_with_keys(self):
    ...

  @classmethod
  def tree_unflatten(cls, aux_data, flat_tree):
    ...

  def __getitem__(self, key: str | int) -> Any:
    """Retrieves the value associated with the given key in the metadata tree.

    If the container is a dict, the key should be a dict key. If the container
    is a list or tuple, the key should be an integer index.

    Args:
      key: The key to retrieve.

    Returns:
      The value associated with the given key.
    """
    ...

  def __contains__(self, key: str | int) -> bool:
    """Checks if the given key is present in the metadata tree.

    If the container is a dict, the key should be a dict key. If the container
    is a list or tuple, the key should be an integer index.

    Args:
      key: The key to check.

    Returns:
      True if the key is present in the tree, False otherwise.
    """
    ...

  def __len__(self) -> int:
    ...

  def __iter__(self):
    ...

  def get(self, key: str, default=None):
    ...

  def keys(self):
    ...

  def values(self):
    ...

  def items(self):
    ...

  @classmethod
  def build(
      cls,
      tree: PyTree,
      *,
      custom_metadata: PyTree | None = None,
  ) -> TreeMetadata:
    """Builds the TreeMetadata."""
    ...


@jax.tree_util.register_pytree_with_keys_class
class _TreeMetadataImpl(TreeMetadata):
  """Default implementation of `TreeMetadata`.

  See parent class for documentation.
  """

  def __init__(
      self,
      *,
      tree: PyTree,
      custom_metadata: PyTree | None = None,
  ):
    self._tree = tree
    self._custom_metadata = custom_metadata
    self._validate_tree_type(tree)

  def _validate_tree_type(self, tree: PyTree):
    # Note: NamedTuple is a subclass of tuple.
    if not isinstance(tree, (dict, list, tuple)):
      raise ValueError(f'Unsupported tree type: {type(tree)}')

  def __repr__(self):
    properties_repr = ''.join(
        [f'  {k}={v}\n' for k, v in self._properties().items()]
    )
    return f'TreeMetadata(\n{properties_repr})'

  @property
  def tree(self) -> PyTree:
    return self._tree

  @property
  def custom_metadata(self) -> PyTree | None:
    return self._custom_metadata


  def tree_flatten(self):
    flat_with_keys, aux_data = self.tree_flatten_with_keys()
    if flat_with_keys:
      _, tree_values = zip(*flat_with_keys)
    else:
      tree_values = []
    return (tree_values, aux_data)

  def tree_flatten_with_keys(self):
    # NOTE: jax.tree_util.tree_flatten_with_path() returns keys that include
    # all levels of nesting. However, the keys expected to be returned by this
    # function are only the top-level keys (sub-tree keys are computed
    # recursively).
    if isinstance(self._tree, dict):
      tree_keys = [jax.tree_util.DictKey(k) for k in self._tree.keys()]
      tree_values = self._tree.values()
    elif tree_utils.isinstance_of_namedtuple(self._tree):
      tree_keys = [jax.tree_util.DictKey(k) for k in self._tree._fields]
      tree_values = [getattr(self._tree, k) for k in self._tree._fields]
    elif isinstance(self._tree, (list, tuple)):
      tree_keys = [jax.tree_util.SequenceKey(i) for i in range(len(self._tree))]
      tree_values = self._tree
    else:
      raise ValueError(f'Unsupported tree type: {type(self._tree)}')

    return (
        list(zip(tree_keys, tree_values)),
        dict(
            tree_type=type(self._tree),
            tree_keys=tree_keys,
            **self._properties(include_tree=False),
        ),
    )

  @classmethod
  def tree_unflatten(cls, aux_data, flat_tree):
    # Pop off any additional `aux_data` fields not from non-tree-
    # mappable properties. Remaining `aux_data` keys are used to initialize
    # `TreeMetadata` properties.
    tree_type = aux_data.pop('tree_type')
    tree_keys = aux_data.pop('tree_keys')
    # `flat_tree` is not truly flat, but is just keys and values for the
    # top-level tree.
    if issubclass(tree_type, dict) or tree_utils.issubclass_of_namedtuple(
        tree_type
    ):
      tree = tree_type(**{
          tree_utils.get_key_name(k): v for k, v in zip(tree_keys, flat_tree)
      })
    elif issubclass(tree_type, (list, tuple)):
      tree = tree_type(flat_tree)
    else:
      raise ValueError(f'Unsupported tree type: {tree_type}')

    return cls(tree=tree, **aux_data)

  def __getitem__(self, key: str | int) -> Any:
    """Retrieves the value associated with the given key in the metadata tree.

    If the container is a dict, the key should be a dict key. If the container
    is a list or tuple, the key should be an integer index.

    Args:
      key: The key to retrieve.

    Returns:
      The value associated with the given key.
    """
    return self.tree[key]

  def __contains__(self, key: str | int) -> bool:
    """Checks if the given key is present in the metadata tree.

    If the container is a dict, the key should be a dict key. If the container
    is a list or tuple, the key should be an integer index.

    Args:
      key: The key to check.

    Returns:
      True if the key is present in the tree, False otherwise.
    """
    return key in self.tree

  def __len__(self) -> int:
    return len(self.tree)

  def __iter__(self):
    return iter(self.tree)

  def get(self, key: str, default=None):
    try:
      return self.__getitem__(key)
    except KeyError:
      return default
    except IndexError:
      return default

  def keys(self):
    return self.tree.keys()

  def values(self):
    return self.tree.values()

  def items(self):
    return self.tree.items()

  def _properties(self, *, include_tree: bool = True) -> dict[str, Any]:
    result = {
        name: getattr(self, name)
        for name, member in inspect.getmembers(type(self))
        if isinstance(member, property)
    }
    if not include_tree:
      result.pop('tree')
    return result

  @classmethod
  def build(
      cls,
      tree: PyTree,
      *,
      custom_metadata: PyTree | None = None,
  ) -> TreeMetadata:
    """Builds the TreeMetadata."""
    return cls(
        tree=tree,
        custom_metadata=custom_metadata,
    )


def build_default_tree_metadata(
    tree: PyTree,
    *,
    custom_metadata: PyTree | None = None,
) -> TreeMetadata:
  """Builds the TreeMetadata using a default implementation."""
  return _TreeMetadataImpl.build(
      tree,
      custom_metadata=custom_metadata,
  )



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage for `array_metadata.ArrayMetadata` (not value.ArrayMetadata)."""

import asyncio
import json
import threading
import time
from typing import Any, Iterator, List, Sequence, Tuple
from absl import logging
from etils import epath
import jax
from orbax.checkpoint._src.metadata import array_metadata as array_metadata_lib
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import types


class PathResolver:
  """Resolves paths for the ArrayMetadata store read and write."""

  _metadata_subdir = 'array_metadatas'

  def _file_name(self, process_index: int | str) -> str:
    return f'process_{process_index}'

  def get_process_index(self, file_path: epath.Path) -> int:
    """Returns the process index from the file path."""
    process_index = file_path.name.removeprefix('process_')
    if process_index.isdigit():
      return int(process_index)
    raise ValueError(
        f'Invalid ArrayMetadata file path: {file_path}; expected file name'
        ' to start with "process_"'
    )

  def get_write_file_path(
      self, checkpoint_dir: epath.Path, process_index: int
  ) -> epath.Path:
    """Returns the file path to write."""
    return (
        checkpoint_dir / self._metadata_subdir / self._file_name(process_index)
    )

  def get_read_file_paths(
      self, checkpoint_dir: epath.Path, process_index: int | None = None
  ) -> Iterator[epath.Path] | epath.Path | None:
    """Returns the file paths to read.

    Args:
      checkpoint_dir: The base path containing metadata for each process.
      process_index: The process index to read. If None, then read all processes
        under `checkpoint_dir`.

    Returns:
      Iterator of file paths to read if `process_index` is None. A file path to
      read if `process_index` is not None. None if `process_index` is not None
      but metadata file does not exist.
    """
    if process_index is None:
      file_name_pattern = self._file_name('*')
      return checkpoint_dir.glob(f'{self._metadata_subdir}/{file_name_pattern}')
    file_path = (
        checkpoint_dir / self._metadata_subdir / self._file_name(process_index)
    )
    if file_path.exists():
      return file_path
    return None


class SerDeserializer:
  """Serializes and deserializes `array_metadata.ArrayMetadata`."""

  def _to_dict(
      self, array_metadata: array_metadata_lib.ArrayMetadata
  ) -> dict[str, Any]:
    """Converts `array_metadata` to a dictionary."""
    return {
        'array_metadata': {
            'param_name': array_metadata.param_name,
            'write_shape': array_metadata.write_shape,
            'chunk_shape': array_metadata.chunk_shape,
            'ext_metadata': array_metadata.ext_metadata,
        }
    }

  def _from_dict(self, obj: dict[str, Any]) -> Any:
    """Converts a json object to `SerializedArrayMetadata` or `obj`."""
    if 'array_metadata' in obj:
      array_metadata = obj['array_metadata']
      return array_metadata_lib.SerializedArrayMetadata(
          param_name=array_metadata['param_name'],
          write_shape=tuple(array_metadata['write_shape']),
          chunk_shape=tuple(array_metadata['chunk_shape']),
          ext_metadata=array_metadata.get('ext_metadata'),
      )
    return obj

  def serialize(
      self, array_metadatas: Sequence[array_metadata_lib.ArrayMetadata]
  ) -> str:
    """Serializes `array_metadatas` to string."""
    obj = {
        'array_metadatas': [
            self._to_dict(array_metadata) for array_metadata in array_metadatas
        ]
    }
    return json.dumps(obj)

  def deserialize(
      self, serialized: str
  ) -> List[array_metadata_lib.SerializedArrayMetadata]:
    """Deserializes `serialized` to `array_metadata.ArrayMetadata`."""
    obj = json.loads(serialized, object_hook=self._from_dict)
    return obj['array_metadatas']


class Store:
  """Storage for `array_metadata.ArrayMetadata` (not value.ArrayMetadata)."""

  def __init__(
      self,
      path_resolver: PathResolver = PathResolver(),
      ser_deser: SerDeserializer = SerDeserializer(),
      primary_host: int | None = 0,  # None means all hosts are primary hosts.
      write_timeout_secs: int = 600,  # 10 minutes.
  ):
    self._path_resolver = path_resolver
    self._ser_deser = ser_deser
    self._primary_host = primary_host
    self._write_timeout_secs = write_timeout_secs

  def set_primary_host(self, primary_host: int | None) -> None:
    """Sets the primary host."""
    self._primary_host = primary_host

  async def _maybe_create_base_dir(self, base_dir: epath.Path) -> None:
    """Primary host creates the base directory, rest of the hosts wait."""
    if multihost.is_primary_host(self._primary_host):
      # primary host creates, rest of the hosts wait.
      return await asyncio.to_thread(
          base_dir.mkdir, parents=True, exist_ok=True
      )

    # non-primary host waits for primary host to create the base dir/folder.
    async def wait_for_base_dir_creation():
      while not await asyncio.to_thread(base_dir.exists):
        await asyncio.sleep(0.25)

    try:
      await asyncio.wait_for(
          wait_for_base_dir_creation(), timeout=self._write_timeout_secs
      )
    except asyncio.TimeoutError as e:
      primary_process = (
          'LOCAL' if self._primary_host is None else self._primary_host
      )
      raise ValueError(
          f'[process_index={multihost.process_index()}] Timed out waiting for'
          f' array_metadatas base directory creation: {base_dir}.'
          f' timeout={self._write_timeout_secs} seconds.'
          f' primary_process={primary_process}'
      ) from e

  async def write(
      self,
      checkpoint_dir: epath.Path,
      array_metadatas: Sequence[array_metadata_lib.ArrayMetadata],
      process_index: int,
  ) -> None:
    """Writes `array_metadatas` to a file under `checkpoint_dir`.

    See `PathResolver.get_write_file_path()` for the file path resolution.

    Args:
      checkpoint_dir: The base path containing metadata for each process.
      array_metadatas: The sequence of metadata to write.
      process_index: The Jax process index used to resolve the file path.
    """
    file_path = self._path_resolver.get_write_file_path(
        checkpoint_dir, process_index
    )
    await self._maybe_create_base_dir(file_path.parent)
    await asyncio.to_thread(
        file_path.write_text, self._ser_deser.serialize(array_metadatas)
    )
    logging.info(
        '[process=%s][thread=%s] Wrote %d array_metadata.ArrayMetadata to %s',
        multihost.process_index(),
        threading.current_thread().name,
        len(array_metadatas),
        file_path,
    )

  async def _get_array_metadatas(
      self,
      array_metadatas_file_path: epath.Path,
  ) -> Tuple[epath.Path, list[array_metadata_lib.SerializedArrayMetadata]]:
    serialized = await asyncio.to_thread(array_metadatas_file_path.read_text)
    return array_metadatas_file_path, self._ser_deser.deserialize(serialized)

  async def read(
      self,
      checkpoint_dir: epath.Path,
      process_index: int | None = None,
  ) -> (
      dict[int, List[array_metadata_lib.SerializedArrayMetadata]]
      | List[array_metadata_lib.SerializedArrayMetadata]
      | None
  ):
    """Reads `SerializedArrayMetadata` from storage under `checkpoint_dir`.

    Args:
      checkpoint_dir: The base path containing metadata for each process.
      process_index: The process index to read. If None, then read all processes
        under `checkpoint_dir`.

    Returns:
      A dictionary of process index to list of metadata if `process_index`
      is None. A list of metadata if `process_index` is not None. None if
      metadata does not exist.
    """
    if not await asyncio.to_thread(checkpoint_dir.exists):
      raise ValueError(
          f'Checkpoint directory does not exist: {checkpoint_dir}.'
      )
    start_time = time.time()
    file_paths = self._path_resolver.get_read_file_paths(
        checkpoint_dir, process_index
    )
    if file_paths is None:
      logging.warning(
          '[process=%s][thread=%s] No metadata found for process_index=%s,'
          ' checkpoint_dir=%s. If the checkpoint does not contain jax.Array'
          ' then it is expected. If checkpoint contains jax.Array then it'
          ' should lead to an error eventually; if no error is raised then it'
          ' is a bug.',
          multihost.process_index(),
          threading.current_thread().name,
          process_index,
          checkpoint_dir,
      )
      return None

    if isinstance(file_paths, epath.Path):
      _, result = await self._get_array_metadatas(file_paths)
      logging.vlog(
          1,
          '[process=%s][thread=%s] Read %s metadata from metadata path=%s'
          ' in %s seconds.',
          multihost.process_index(),
          threading.current_thread().name,
          len(result),
          file_paths,
          time.time() - start_time,
      )
      return result

    getter_coros = []
    for file_path in file_paths:
      getter_coros.append(self._get_array_metadatas(file_path))
    path_metadatas_pairs = await asyncio.gather(*getter_coros)
    result = {
        self._path_resolver.get_process_index(file_path): metadatas
        for file_path, metadatas in path_metadatas_pairs
    }
    if not result:
      logging.warning(
          '[process=%s][thread=%s] No metadata found for any process_index,'
          ' checkpoint_dir=%s. time elapsed=%s seconds. If the checkpoint does'
          ' not contain jax.Array then it is expected. If checkpoint contains'
          ' jax.Array then it should lead to an error eventually; if no error'
          ' is raised then it is a bug.',
          multihost.process_index(),
          threading.current_thread().name,
          checkpoint_dir,
          time.time() - start_time,
      )
      return None

    logging.vlog(
        1,
        '[process=%s][thread=%s] Read all metadata from checkpoint_dir=%s in %s'
        ' seconds.',
        multihost.process_index(),
        threading.current_thread().name,
        checkpoint_dir,
        time.time() - start_time,
    )
    return result


def resolve_array_metadata_store(
    type_handler_registry: types.TypeHandlerRegistry,
) -> Store | None:
  """Returns the ArrayMetadata Store or None.

  Gets ArrayMetadata Store from TypeHandler for jax.Array. ArrayMetadata
  persistence is only supported for jax.Array.

  Args:
    type_handler_registry: TypeHandlerRegistry to search TypeHandler for
      jax.Array.
  """
  try:
    array_handler = type_handler_registry.get(jax.Array)
  except ValueError:
    logging.warning(
        'No jax.Array in TypeHandlerRegistry: ArrayMetadata persistence is'
        ' disabled.'
    )
    return None
  else:
    if hasattr(array_handler, '_array_metadata_store'):
      return array_handler._array_metadata_store  # pylint: disable=protected-access
    else:
      logging.warning(
          'No ArrayMetadata Store defined in TypeHandler for jax.Array: %s!'
          ' Subchunking is disabled.',
          array_handler.__class__.__qualname__,
      )
      return None


class Validator:
  """Validates ArrayMetadata."""

  def validate_all_array_metadatas(
      self,
      array_metadatas: dict[
          int, List[array_metadata_lib.SerializedArrayMetadata]
      ],
  ) -> None:
    """Validates that all processes have the same array metadatas.

    Args:
      array_metadatas: A dictionary of process index to list of metadata.
    """
    start_time = time.time()
    ref_process_index, ref_process_array_metadatas = next(
        iter(array_metadatas.items())
    )
    if not ref_process_array_metadatas:
      raise ValueError(
          'ArrayMetadata Store contains no metadata for process_index='
          f'{ref_process_index}.'
      )
    if len(array_metadatas) == 1:
      # check if the number of processes are indeed just one.
      self._validate_process_count(ref_process_index=ref_process_index)
      logging.warning(
          '[process=%s][thread=%s] Skipped cross-host ArrayMetadata validation'
          ' because only one process is found: process_index=%s.',
          multihost.process_index(),
          threading.current_thread().name,
          ref_process_index,
      )
      return

    ref_process_cache = {
        array_metadata.param_name: array_metadata
        for array_metadata in ref_process_array_metadatas
    }
    for process_index, process_array_metadatas in array_metadatas.items():
      if process_index == ref_process_index:
        continue
      process_cache = {
          array_metadata.param_name: array_metadata
          for array_metadata in process_array_metadatas
      }
      # Check if the number of params is the same.
      self._validate_param_count(
          ref_process_index=ref_process_index,
          ref_process_array_metadatas=ref_process_array_metadatas,
          ref_process_cache=ref_process_cache,
          process_index=process_index,
          process_array_metadatas=process_array_metadatas,
          process_cache=process_cache,
      )
      # Check if the params are the same.
      self._validate_params(
          ref_process_index=ref_process_index,
          ref_process_cache=ref_process_cache,
          process_index=process_index,
          process_cache=process_cache,
      )
      # Check if the chunk_shape, write_shape and ext_metadata are the same for
      # each param.
      self._validate_metadata(
          ref_process_index=ref_process_index,
          ref_process_cache=ref_process_cache,
          process_index=process_index,
          process_cache=process_cache,
      )
    logging.info(
        '[process=%s][thread=%s] Validated ArrayMetadata from all %s hosts in'
        ' %s seconds.',
        multihost.process_index(),
        threading.current_thread().name,
        len(array_metadatas),
        time.time() - start_time,
    )

  def _validate_process_count(
      self,
      *,
      ref_process_index: int,
  ) -> None:
    """Validates that the number of processes is just one."""
    process_count = multihost.process_count()
    if process_count != 1:
      raise ValueError(
          'ArrayMetadata Store contains metadata from just one process'
          f' (process_index={ref_process_index}), but found'
          f' {process_count} processes.'
      )

  def _validate_param_count(
      self,
      *,
      ref_process_index: int,
      ref_process_array_metadatas: List[
          array_metadata_lib.SerializedArrayMetadata
      ],
      ref_process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
      process_index: int,
      process_array_metadatas: List[array_metadata_lib.SerializedArrayMetadata],
      process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
  ) -> None:
    """Validates that the number of params is the same."""
    if len(ref_process_array_metadatas) != len(process_array_metadatas):
      missing_in_process = ref_process_cache.keys() - process_cache.keys()
      missing_in_ref_process = process_cache.keys() - ref_process_cache.keys()
      diff_msg = 'Diff:'
      if missing_in_process:
        diff_msg += (
            f' process_index={process_index} is missing'
            f' {len(missing_in_process)} params: {missing_in_process}.'
        )
      if missing_in_ref_process:
        diff_msg += (
            f' process_index={ref_process_index} is missing'
            f' {len(missing_in_ref_process)} params:'
            f' {missing_in_ref_process}.'
        )
      raise ValueError(
          'ArrayMetadata Store contains different number of params:'
          f' process_index={ref_process_index} has'
          f' {len(ref_process_array_metadatas)}, but'
          f' process_index={process_index} has'
          f' {len(process_array_metadatas)} params. {diff_msg}'
      )

  def _validate_params(
      self,
      *,
      ref_process_index: int,
      ref_process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
      process_index: int,
      process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
  ) -> None:
    """Validates that the params are the same."""
    missing_in_process = ref_process_cache.keys() - process_cache.keys()
    if missing_in_process:
      raise ValueError(
          'ArrayMetadata Store contains different params: comparing with'
          f' process_index={ref_process_index},'
          f' process_index={process_index} is missing'
          f' {len(missing_in_process)} params: {missing_in_process}.'
      )
    missing_in_ref_process = process_cache.keys() - ref_process_cache.keys()
    if missing_in_ref_process:
      raise ValueError(
          'ArrayMetadata Store contains different params: comparing with'
          f' process_index={process_index},'
          f' process_index={ref_process_index} is missing'
          f' {len(missing_in_ref_process)} params: {missing_in_ref_process}.'
      )

  def _validate_metadata(
      self,
      *,
      ref_process_index: int,
      ref_process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
      process_index: int,
      process_cache: dict[str, array_metadata_lib.SerializedArrayMetadata],
  ) -> None:
    """Validates that chunk_shape and write_shape are the same for a param."""
    for param_name, array_metadata in process_cache.items():
      ref_array_metadata = ref_process_cache[param_name]
      if array_metadata.chunk_shape != ref_array_metadata.chunk_shape:
        raise ValueError(
            'ArrayMetadata Store contains different chunk_shape for param:'
            f' {param_name}. process_index={process_index} has'
            f' {array_metadata.chunk_shape}, but'
            f' process_index={ref_process_index} has'
            f' {ref_array_metadata.chunk_shape}.'
        )
      if array_metadata.write_shape != ref_array_metadata.write_shape:
        raise ValueError(
            'ArrayMetadata Store contains different write_shape for param:'
            f' {param_name}. process_index={process_index} has'
            f' {array_metadata.write_shape}, but'
            f' process_index={ref_process_index} has'
            f' {ref_array_metadata.write_shape}.'
        )

      # Check if the RANDOM_KEY_IMPL is the same.
      if not isinstance(
          array_metadata.ext_metadata, type(ref_array_metadata.ext_metadata)
      ):
        raise ValueError(
            'Different ext_metadata types param:'
            f' {param_name}. process_index={process_index} has'
            f' {array_metadata.ext_metadata}, but'
            f' process_index={ref_process_index} has'
            f' {ref_array_metadata.ext_metadata}.'
        )

      if (
          isinstance(ref_array_metadata.ext_metadata, dict)
          and isinstance(array_metadata.ext_metadata, dict)
      ) and (
          (
              ref_key := ref_array_metadata.ext_metadata.get(
                  array_metadata_lib.RANDOM_KEY_IMPL
              )
          )
          != (
              arr_key := array_metadata.ext_metadata.get(
                  array_metadata_lib.RANDOM_KEY_IMPL
              )
          )
      ):
        raise ValueError(
            'ArrayMetadata Store contains different'
            f' ext_metadata.{array_metadata_lib.RANDOM_KEY_IMPL} for param:'
            f' {param_name}. process_index={process_index} has {ref_key}, but'
            f' process_index={ref_process_index} has {arr_key}.'
        )


async def validate_all_array_metadatas(
    validator: Validator,
    array_metadata_store: Store,
    directory: epath.Path,
) -> None:
  """Validates that all processes have the same array metadatas.

  Args:
    validator: The `Validator` instance for validation.
    array_metadata_store: The `Store` instance for reading metadata from
      `directory`.
    directory: The checkpoint directory with array_metadatas subdir.
  """
  array_metadatas = await array_metadata_store.read(directory)
  if array_metadatas is not None:
    assert isinstance(array_metadatas, dict)  # read all processes.
    validator.validate_all_array_metadatas(array_metadatas)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class to manage checkpoints: AbstractCheckpointManager."""

import abc
from typing import Any, Mapping, Optional, Protocol, Sequence, Union

from etils import epath
from orbax.checkpoint import args as args_lib
from orbax.checkpoint._src.metadata import checkpoint

PyTree = Any
SaveParams = Mapping[str, Any]
RestoreParams = SaveParams


class AbstractCheckpointManager(Protocol):
  """Interface to manage checkpoints.

  Allows a user to save and restore objects for which a Checkpointer
  implementation exists (e.g. PyTreeCheckpointer for PyTrees). The class
  keeps track of multiple checkpointable objects in the following structure::

    path/to/directory/    (top-level directory)
      0/    (step)
        params/    (first saveable)
          ...
        metadata/    (second saveable)
          ...
      1/    (step)
        ...
      2/    (step)
        ...
      ...
  """

  @property
  @abc.abstractmethod
  def directory(self) -> epath.Path:
    """Returns the top-level directory containing checkpoints for all items."""

  @abc.abstractmethod
  def all_steps(self, read: bool = False) -> Sequence[int]:
    """Returns all steps tracked by the manager.

    Args:
      read: If True, forces a read directly from the storage location.
        Otherwise, a cached result can be returned.

    Returns:
      A sequence of steps (integers)
    """

  @abc.abstractmethod
  def latest_step(self) -> Optional[int]:
    """Returns the latest step saved.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """

  @abc.abstractmethod
  def best_step(self) -> Optional[int]:
    """Returns the best step saved, as defined by `options.best_fn`.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """

  @abc.abstractmethod
  def reload(self):
    """Reloads internal properties.

    Resets internal cache of checkpoint steps, in case the directory managed
    by this object has been updated externally.
    """

  @abc.abstractmethod
  def reached_preemption(self, step: int) -> bool:
    """Returns True if a preemption sync point has been reached."""

  @abc.abstractmethod
  def should_save(self, step: int) -> bool:
    """Returns True if a checkpoint should be saved for the current step.

    This depends the previous step and save interval.

    Args:
      step: int

    Returns:
      True if the checkpoint should be saved.
    """

  @abc.abstractmethod
  def delete(self, step: int):
    """Deletes a step checkpoint."""

  @abc.abstractmethod
  def save(
      self,
      step: int,
      *args,
      **kwargs,
  ) -> bool:
    """Saves the given step."""

  @abc.abstractmethod
  def restore(
      self,
      step: Optional[int],
      *args,
      **kwargs,
  ) -> Union[Any, Mapping[str, Any], args_lib.Composite]:
    """Restores the given step."""

  @abc.abstractmethod
  def item_metadata(
      self, step: int
  ) -> Union[Any, Mapping[str, Any], args_lib.Composite]:
    """Returns metadata for all known items."""

  @abc.abstractmethod
  def metadata(
      self, step: int | None = None,
  ) -> checkpoint.StepMetadata | checkpoint.RootMetadata:
    """Returns `StepMetadata` for the specified step, or `RootMetadata` all.

    If step is specified, only return `StepMetadata` for that step.
    Otherwise, return `RootMetadata`.

    Args:
      step: Step for which to retrieve `StepMetadata`. If None, returns
        `RootMetadata`.

    Returns:
      Metadata for the specified step (`StepMetadata`), or all steps
      (`RootMetadata`).
    """

  @abc.abstractmethod
  def metrics(self, step: int) -> Optional[PyTree]:
    """Returns metrics for step, if present."""

  @abc.abstractmethod
  def wait_until_finished(self):
    """Blocks until any incomplete save operations are completed.

    Note that this method will typically be a no-op if all checkpointers are
    synchronous, since old checkpoints are already cleaned up immediately after
    completing `save`, and there is no background thread to wait for.

    If some checkpointers are of type AsyncCheckpointer, however, this method
    will wait until each of these checkpointers is finished.
    """

  @abc.abstractmethod
  def check_for_errors(self):
    """Checks for any outstanding errors in completed asynchronous save operations.

    Delegates to underlying Checkpointer.
    """

  @abc.abstractmethod
  def close(self):
    """Waits for outstanding operations to finish and closes Checkpointers."""



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for saving."""

# pylint: disable=protected-access

import threading

from etils import epath
import jax
import numpy as np
import orbax.checkpoint as ocp
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.context import options as options_lib
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


def _make_v0_save_args(
    pytree: tree_types.PyTree,
    create_array_storage_options_fn: (
        options_lib.CreateArrayStorageOptionsFn | None
    ),
) -> tree_types.PyTree:
  """Creates v0-compatible `SaveArgs` from v1 storage options."""
  if create_array_storage_options_fn is None:
    save_args = None
  else:
    array_storage_options_pytree = jax.tree.map_with_path(
        create_array_storage_options_fn, pytree
    )
    save_args = jax.tree.map(
        lambda v: ocp.SaveArgs(
            dtype=np.dtype(v.dtype),
            chunk_byte_size=v.chunk_byte_size,
            shard_axes=v.shard_axes,
        ),
        array_storage_options_pytree,
    )
  return save_args


def _get_concurrent_gb(concurrent_bytes: int | None) -> int | None:
  if concurrent_bytes:
    return max(int(concurrent_bytes / 1e9), 1)
  return None


def save_pytree(
    directory: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    # Generic arguments.
    force: bool = False,
    custom_metadata: ocp.tree.JsonType | None = None,
):
  """Saves a PyTree.

  The operation blocks until complete. For improved performance, consider using
  `save_async` instead.

  Args:
    directory: The directory to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    force: Whether to allow the save to proceed even if it would fully overwrite
      an existing checkpoint.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """
  context = context_lib.get_context()

  handler_registry = ocp.handlers.create_default_handler_registry(
      pytree=ocp.PyTreeCheckpointHandler(
          use_ocdbt=context.pytree_options.use_ocdbt,
          use_zarr3=context.pytree_options.use_zarr3,
          save_concurrent_gb=_get_concurrent_gb(
              context.pytree_options.save_concurrent_bytes
          ),
      )
  )
  ckptr = ocp.Checkpointer(
      ocp.CompositeCheckpointHandler(handler_registry=handler_registry)
  )
  save_args = _make_v0_save_args(
      pytree, context.pytree_options.create_array_storage_options_fn
  )
  args = ocp.args.Composite(
      pytree=ocp.args.PyTreeSave(
          pytree,
          save_args=save_args,
          ocdbt_target_data_file_size=(
              context.pytree_options.ocdbt_target_data_file_size
          ),
          enable_pinned_host_transfer=(
              context.pytree_options.enable_pinned_host_transfer
          ),
      )
  )
  ckptr.save(directory, args=args, force=force, custom_metadata=custom_metadata)


class _SaveResponse(async_types.AsyncResponse[None]):
  """An `AsyncResponse` representing the result of `save_pytree_async`.

  TODO(cpgaffney): Note that a memory leak is possible if the user does not
  call `result`.
  """

  def __init__(self, checkpointer: ocp.AsyncCheckpointer):
    self._checkpointer = checkpointer
    self._thread = threading.Thread(target=self._wait_for_save)
    self._thread.start()

  def _wait_for_save(self):
    self._checkpointer.wait_until_finished()

  def result(self, timeout: float | None = None) -> None:
    self._thread.join()
    self._checkpointer.close()


def save_pytree_async(
    directory: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    # Generic arguments.
    force: bool = False,
    custom_metadata: ocp.tree.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Saves a PyTree asynchronously.

  Unlike `save`, this function returns immediately after the save operation is
  scheduled (except for certain operations, like device-to-host copying of
  on-device arrays, which must happen on the main thread). Further writing
  operations continue in a background thread. An `AsyncResponse` is returned
  that can be used to block until the save is complete (using
  `response.result()`). Make sure to wait for completion before attempting to
  load the checkpoint or exiting the program.

  Args:
    directory: The directory to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    force: Whether to allow the save to proceed even if it would fully overwrite
      an existing checkpoint.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  context = context_lib.get_context()

  handler_registry = ocp.handlers.create_default_handler_registry(
      pytree=ocp.PyTreeCheckpointHandler(
          use_ocdbt=context.pytree_options.use_ocdbt,
          use_zarr3=context.pytree_options.use_zarr3,
          save_concurrent_gb=_get_concurrent_gb(
              context.pytree_options.save_concurrent_bytes
          ),
      )
  )
  ckptr = ocp.AsyncCheckpointer(
      ocp.CompositeCheckpointHandler(handler_registry=handler_registry)
  )
  save_args = _make_v0_save_args(
      pytree, context.pytree_options.create_array_storage_options_fn
  )
  args = ocp.args.Composite(
      pytree=ocp.args.PyTreeSave(
          pytree,
          save_args=save_args,
          ocdbt_target_data_file_size=(
              context.pytree_options.ocdbt_target_data_file_size
          ),
          enable_pinned_host_transfer=(
              context.pytree_options.enable_pinned_host_transfer
          ),
      )
  )
  directory = epath.Path(directory)
  ckptr.save(directory, args=args, force=force, custom_metadata=custom_metadata)
  return _SaveResponse(ckptr)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for debugging local checkpoints with missing chunks."""

import asyncio
from typing import Any
from absl import logging
from etils import epath
import jax
from orbax.checkpoint._src.arrays import abstract_arrays
from orbax.checkpoint._src.arrays import types
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.serialization import types as serialization_types
from orbax.checkpoint._src.tree import utils as tree_utils
import tensorstore as ts


Index = types.Index
Shape = types.Shape
ChunkId = tuple[int, ...]
PyTree = Any


def index_to_chunk_id(
    idx: Index, global_shape: Shape, shard_shape: Shape
) -> ChunkId:
  """Converts index to chunk id."""
  assert len(idx) == len(global_shape) == len(shard_shape)
  chunks = tuple([gs / ss for gs, ss in zip(global_shape, shard_shape)])
  chunk_id = [-1] * len(chunks)
  for d, sl in enumerate(idx):
    start = sl.start
    stop = sl.stop
    if start is None:
      start = 0
    if stop is None:
      stop = global_shape[d]
    assert sl.step == 1 or sl.step is None
    assert start % shard_shape[d] == 0 and stop % shard_shape[d] == 0
    assert stop - start == shard_shape[d]
    chunk_id[d] = start // shard_shape[d]
  return tuple(chunk_id)


def chunk_id_to_index(chunk_id: ChunkId, shard_shape: Shape) -> Index:
  """Converts chunk id to index."""
  assert len(chunk_id) == len(shard_shape)
  idx = []
  for d, dim_id in enumerate(chunk_id):
    start = dim_id * shard_shape[d]
    stop = start + shard_shape[d]
    idx.append(slice(start, stop))
  assert len(idx) == len(shard_shape)
  return tuple(idx)


async def get_chunk_ids_from_tensorstore(
    t: ts.TensorStore, *, use_zarr3: bool, shape: Shape | None = None
) -> list[ChunkId]:
  """List chunks available in the TensorStore KVStore."""
  separator = '/' if use_zarr3 else '.'
  kvstore = await t.kvstore.list()
  kvstore = [v.decode('utf-8') for v in kvstore]
  result = [
      tuple([int(x) for x in v.split(separator) if x != 'c'])
      for v in kvstore
      if 'zarr' not in v
  ]
  if shape is not None:
    for chunk_id in result:
      assert len(chunk_id) == len(shape)
  return result


async def open_tensorstore(
    directory: epath.Path,
    param_name: str,
    *,
    use_ocdbt: bool,
    use_zarr3: bool,
) -> ts.TensorStore:
  """Opens TensorStore for the given parameter for reading."""
  info = serialization_types.ParamInfo(
      name=param_name,
      path=directory / param_name,
      parent_dir=directory,
      is_ocdbt_checkpoint=use_ocdbt,
      use_zarr3=use_zarr3,
      ts_context=type_handlers.get_ts_context(use_ocdbt=use_ocdbt),
  )
  tspec = type_handlers.get_json_tspec_read(info, use_ocdbt=use_ocdbt)
  return await ts.open(
      ts.Spec(tspec),
      read=True,
      write=False,
      create=False,
      open=True,
      context=info.ts_context,
  )


async def get_present_and_missing_chunks(
    directory: epath.Path,
    param_name: str,
    array: jax.Array | jax.ShapeDtypeStruct,
    *,
    use_ocdbt: bool,
    use_zarr3: bool,
) -> tuple[list[ChunkId], list[ChunkId]]:
  """From TensorStore, get present and missing chunk ids."""
  t = await open_tensorstore(
      directory, param_name, use_ocdbt=use_ocdbt, use_zarr3=use_zarr3
  )
  present_chunk_ids = await get_chunk_ids_from_tensorstore(
      t, use_zarr3=use_zarr3, shape=array.shape
  )
  chunk_ids_set = set(present_chunk_ids)
  missing_chunk_ids = []
  for d, idx in array.sharding.devices_indices_map(array.shape).items():
    if d not in jax.local_devices():  # non-local indices not expected
      continue
    chunk_id = index_to_chunk_id(
        idx, array.shape, array.sharding.shard_shape(array.shape)
    )
    if chunk_id not in chunk_ids_set:
      missing_chunk_ids.append(chunk_id)
  return present_chunk_ids, missing_chunk_ids


async def print_chunk_debug_info(
    directory: epath.Path,
    restore_args: PyTree | None = None,
):
  """Prints debug info before restoring the parameter."""
  arrays = jax.tree.map(abstract_arrays.to_shape_dtype_struct, restore_args)
  param_names = tree_utils.get_param_names(arrays, include_empty_nodes=False)
  flat_arrays, _ = jax.tree.flatten(arrays)
  flat_param_names, _ = jax.tree.flatten(param_names)
  logging.vlog(1, 'Found %d arrays.', len(flat_arrays))
  logging.vlog(1, 'Param names to check: %s', flat_param_names)
  assert len(flat_arrays) == len(flat_param_names)

  chunk_infos = await asyncio.gather(*[
      get_present_and_missing_chunks(
          directory, n, a, use_ocdbt=True, use_zarr3=True
      )
      for n, a in zip(flat_param_names, flat_arrays)
  ])

  for arr, param_name, chunk_info in zip(
      flat_arrays, flat_param_names, chunk_infos
  ):
    present_chunk_ids, missing_chunk_ids = chunk_info
    shard_shape = arr.sharding.shard_shape(arr.shape)

    logging.vlog(1, 'Logging present and missing chunks for %s.', param_name)
    if present_chunk_ids:
      logging.info('Present chunks:')
      for chunk_id in present_chunk_ids:
        logging.vlog(
            1,
            '  [present] index: %s -- chunk ID: %s',
            chunk_id_to_index(chunk_id, shard_shape),
            chunk_id,
        )
    if missing_chunk_ids:
      logging.vlog(1, 'Missing chunks:')
      for chunk_id in missing_chunk_ids:
        logging.vlog(
            1,
            '  [missing] index: %s -- chunk ID: %s',
            chunk_id_to_index(chunk_id, shard_shape),
            chunk_id,
        )


def print_devices_indices_debug_info(
    restore_args: PyTree,
):
  """Prints debug info before restoring the parameter."""
  arrays = jax.tree.map(abstract_arrays.to_shape_dtype_struct, restore_args)
  param_names = tree_utils.get_param_names(arrays, include_empty_nodes=False)
  flat_arrays, _ = jax.tree.flatten(arrays)
  flat_param_names, _ = jax.tree.flatten(param_names)
  assert len(flat_arrays) == len(flat_param_names)

  for arr, param_name in zip(flat_arrays, flat_param_names):
    devices_indices_map = arr.sharding.devices_indices_map(arr.shape)
    logging.vlog(1, 'Device -> index map for %s.', param_name)
    for d, idx in devices_indices_map.items():
      logging.vlog(1, '  %s -> %s', d, idx)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CheckpointHandler interface."""

import abc
from typing import Any, Optional
from etils import epath


class CheckpointHandler(abc.ABC):
  """An interface providing save/restore methods used on a savable item.

  Item may be a PyTree, Dataset, or any other supported object.

  NOTE: Users should avoid using CheckpointHandler independently. Use
  Checkpointer or CheckpointManager.
  """

  @abc.abstractmethod
  def save(self, directory: epath.Path, *args, **kwargs):
    """Saves the provided item synchronously.

    Args:
      directory: the directory to save to.
      *args: additional arguments for save.
      **kwargs: additional arguments for save.
    """
    pass

  @abc.abstractmethod
  def restore(self, directory: epath.Path, *args, **kwargs) -> Any:
    """Restores the provided item synchronously.

    Args:
      directory: the directory to restore from.
      *args: additional arguments for restore.
      **kwargs: additional arguments for restore.

    Returns:
      The restored item.
    """
    pass

  def metadata(self, directory: epath.Path) -> Optional[Any]:
    """Returns metadata about the saved item.

    Ideally, this is a cheap way to collect information about the checkpoint
    without requiring a full restoration.

    Args:
      directory: the directory where the checkpoint is located.

    Returns:
      item metadata
    """
    pass

  def finalize(self, directory: epath.Path) -> None:
    """Optional, custom checkpoint finalization callback.

    Will be called on a single worker after all workers have finished writing.

    Args:
      directory: the directory where the checkpoint is located.
    """
    pass

  def close(self):
    """Closes the CheckpointHandler."""
    pass

  @classmethod
  def typestr(cls) -> str:
    """A unique identifier for the CheckpointHandler type."""
    return f"{cls.__module__}.{cls.__qualname__}"



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at step level."""

import dataclasses
from typing import Any

from absl import logging
from etils import epath
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
StepMetadata = checkpoint.StepMetadata
CompositeCheckpointHandlerTypeStrs = (
    checkpoint.CompositeCheckpointHandlerTypeStrs
)
CheckpointHandlerTypeStr = checkpoint.CheckpointHandlerTypeStr
CompositeItemMetadata = checkpoint.CompositeItemMetadata
SingleItemMetadata = checkpoint.SingleItemMetadata
StepStatistics = step_statistics.SaveStepStatistics


def get_step_metadata(path: epath.PathLike) -> StepMetadata:
  """Returns StepMetadata for a given checkpoint directory."""
  metadata_file_path = checkpoint.step_metadata_file_path(path)
  serialized_metadata = checkpoint.metadata_store(enable_write=False).read(
      metadata_file_path
  )
  if serialized_metadata is None:
    raise ValueError(f'Step metadata not found at {metadata_file_path}')
  return deserialize(serialized_metadata)


def serialize(metadata: StepMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""

  # Per item metadata is already saved in the item subdirectories.
  del metadata.item_metadata

  # Save only float performance metrics.
  performance_metrics = metadata.performance_metrics
  float_metrics = {
      metric: val
      for metric, val in dataclasses.asdict(performance_metrics).items()
      if isinstance(val, float)
  }

  return {
      'item_handlers': metadata.item_handlers,
      'metrics': metadata.metrics,
      'performance_metrics': float_metrics,
      'init_timestamp_nsecs': metadata.init_timestamp_nsecs,
      'commit_timestamp_nsecs': metadata.commit_timestamp_nsecs,
      'custom_metadata': metadata.custom_metadata,
  }


# TODO(adamcogdell): Reduce code duplication with deserialize().
def serialize_for_update(**kwargs) -> SerializedMetadata:
  """Validates and serializes `kwargs` to a dictionary.

  To be used with MetadataStore.update().

  Args:
    **kwargs: The kwargs to be serialized.

  Returns:
    A dictionary of the serialized kwargs.
  """
  validated_kwargs = {}

  if 'item_handlers' in kwargs:
    utils.validate_type(kwargs['item_handlers'], [dict, str])
    item_handlers = kwargs.get('item_handlers')
    if isinstance(item_handlers, CompositeCheckpointHandlerTypeStrs):
      for k in kwargs.get('item_handlers'):
        utils.validate_type(k, str)
      validated_kwargs['item_handlers'] = item_handlers
    elif isinstance(item_handlers, CheckpointHandlerTypeStr):
      validated_kwargs['item_handlers'] = item_handlers

  if 'metrics' in kwargs:
    utils.validate_type(kwargs['metrics'], dict)
    for k in kwargs.get('metrics', {}) or {}:
      utils.validate_type(k, str)
    validated_kwargs['metrics'] = kwargs.get('metrics', {})

  if 'performance_metrics' in kwargs:
    utils.validate_type(kwargs['performance_metrics'], [dict, StepStatistics])
    performance_metrics = kwargs.get('performance_metrics', {})
    if isinstance(performance_metrics, StepStatistics):
      performance_metrics = dataclasses.asdict(performance_metrics)
    float_metrics = {
        metric: val
        for metric, val in performance_metrics.items()
        if isinstance(val, float)
    }
    validated_kwargs['performance_metrics'] = float_metrics

  if 'init_timestamp_nsecs' in kwargs:
    utils.validate_type(kwargs['init_timestamp_nsecs'], int)
    validated_kwargs['init_timestamp_nsecs'] = (
        kwargs.get('init_timestamp_nsecs', None)
    )

  if 'commit_timestamp_nsecs' in kwargs:
    utils.validate_type(kwargs['commit_timestamp_nsecs'], int)
    validated_kwargs['commit_timestamp_nsecs'] = (
        kwargs.get('commit_timestamp_nsecs', None)
    )

  if 'custom_metadata' in kwargs:
    if kwargs['custom_metadata'] is None:
      validated_kwargs['custom_metadata'] = {}
    else:
      utils.validate_type(kwargs['custom_metadata'], dict)
      for k in kwargs.get('custom_metadata', {}) or {}:
        utils.validate_type(k, str)
      validated_kwargs['custom_metadata'] = kwargs.get('custom_metadata', {})

  for k in kwargs:
    if k not in validated_kwargs:
      raise ValueError('Provided metadata contains unknown key %s.' % k)

  return validated_kwargs


def deserialize(
    metadata_dict: SerializedMetadata,
    item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
    metrics: dict[str, Any] | None = None,
) -> StepMetadata:
  """Deserializes `metadata_dict` and other kwargs to `StepMetadata`."""
  validated_metadata_dict = {}

  item_handlers = metadata_dict.get('item_handlers', None)
  if item_handlers is not None:
    utils.validate_type(item_handlers, [dict, str])
    if isinstance(item_handlers, CompositeCheckpointHandlerTypeStrs):
      for k in metadata_dict.get('item_handlers', {}) or {}:
        utils.validate_type(k, str)
      validated_metadata_dict['item_handlers'] = item_handlers
    elif isinstance(item_handlers, CheckpointHandlerTypeStr):
      validated_metadata_dict['item_handlers'] = item_handlers
  else:
    validated_metadata_dict['item_handlers'] = None

  if isinstance(item_metadata, CompositeItemMetadata):
    validated_metadata_dict['item_metadata'] = {}
    for k, v in item_metadata.items():
      utils.validate_type(k, str)
      validated_metadata_dict['item_metadata'][k] = v
  else:
    validated_metadata_dict['item_metadata'] = item_metadata

  serialized_metrics = metadata_dict.get('metrics', None)
  if serialized_metrics is not None:
    utils.validate_type(serialized_metrics, dict)
    for k in serialized_metrics:
      utils.validate_type(k, str)
  validated_metadata_dict['metrics'] = serialized_metrics or {}
  if metrics is not None:
    utils.validate_type(metrics, dict)
    for k, v in metrics.items():
      utils.validate_type(k, str)
      validated_metadata_dict['metrics'][k] = v

  performance_metrics = metadata_dict.get('performance_metrics', None)
  if performance_metrics is not None:
    utils.validate_type(performance_metrics, dict)
    for k in performance_metrics:
      utils.validate_type(k, str)
      utils.validate_type(performance_metrics[k], float)
  validated_metadata_dict['performance_metrics'] = StepStatistics(
      **(performance_metrics or {})
  )

  init_timestamp_nsecs = metadata_dict.get('init_timestamp_nsecs', None)
  if init_timestamp_nsecs is not None:
    utils.validate_type(init_timestamp_nsecs, int)
  validated_metadata_dict['init_timestamp_nsecs'] = init_timestamp_nsecs

  commit_timestamp_nsecs = metadata_dict.get('commit_timestamp_nsecs', None)
  if commit_timestamp_nsecs is not None:
    utils.validate_type(commit_timestamp_nsecs, int)
  validated_metadata_dict['commit_timestamp_nsecs'] = commit_timestamp_nsecs

  validated_metadata_dict['custom_metadata'] = (
      utils.validate_and_process_custom_metadata(
          metadata_dict.get('custom_metadata', None)
      )
  )

  for k in metadata_dict:
    if k not in validated_metadata_dict:
      if (
          'custom_metadata' in metadata_dict
          and metadata_dict['custom_metadata']
      ):
        raise ValueError(
            'Provided metadata contains unknown key %s, and the custom_metadata'
            ' field is already defined.' % k
        )
      logging.warning(
          'Provided metadata contains unknown key %s. Adding it to'
          ' custom_metadata.',
          k,
      )
      validated_metadata_dict['custom_metadata'][k] = metadata_dict[k]

  return StepMetadata(**validated_metadata_dict)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for interacting with the atomicity module.

Mainly provides helper functions like `get_default_temporary_path_class` that
help interacting with the `atomicity` module, but can have larger dependencies
than we would want to introduce into the base `atomicity` module.
"""

from typing import Type
from etils import epath
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import atomicity_types
from orbax.checkpoint._src.path import step as step_lib


def get_item_default_temporary_path_class(
    final_path: epath.Path,
) -> Type[atomicity_types.TemporaryPath]:
  """Returns the default temporary path class for a given sub-item path."""
  if step_lib.is_gcs_path(final_path):
    return atomicity.CommitFileTemporaryPath
  else:
    return atomicity.AtomicRenameTemporaryPath


def get_default_temporary_path_class(
    final_path: epath.Path,
) -> Type[atomicity_types.TemporaryPath]:
  """Returns the default temporary path class for a given checkpoint path."""
  if step_lib.is_gcs_path(final_path):
    return atomicity.CommitFileTemporaryPath
  else:
    return atomicity.AtomicRenameTemporaryPath



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for save/load tests."""

# pylint: disable=missing-class-docstring,protected-access,missing-function-docstring

from absl.testing import parameterized
from etils import epath
import flax
import jax
from jax import numpy as jnp
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.tree import utils as tree_utils
import orbax.checkpoint.experimental.v1 as ocp
from orbax.checkpoint.experimental.v1._src.testing import array_utils as array_test_utils
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PyTree = tree_types.PyTree

create_sharded_array = array_test_utils.create_sharded_array
create_numpy_pytree = array_test_utils.create_numpy_pytree
create_sharded_pytree = array_test_utils.create_sharded_pytree
as_abstract_type = array_test_utils.as_abstract_type


class SaveLoadTestBase:

  class Test(parameterized.TestCase):

    def setUp(self):
      super().setUp()

      self.directory = (
          epath.Path(self.create_tempdir(name='saving_test').full_path) / 'ckpt'
      )
      self.pytree, self.abstract_pytree = create_numpy_pytree()
      self.numpy_pytree, self.abstract_numpy_pytree = create_sharded_pytree()

      test_utils.set_tensorstore_driver_for_test()
      test_utils.sync_global_processes('SavingTest:setup_complete')

    def tearDown(self):
      super().tearDown()
      test_utils.sync_global_processes('SavingTest:teardown_complete')

    def save_and_wait(self, *args, use_async: bool, **kwargs):
      if use_async:
        response = ocp.save_pytree_async(*args, **kwargs)
        self.assertIsNotNone(response)
        self.assertIsNone(response.result())
      else:
        self.assertIsNone(ocp.save_pytree(*args, **kwargs))

    def load_and_wait(self, *args, use_async: bool, **kwargs):
      del use_async
      return ocp.load_pytree(*args, **kwargs)

    @parameterized.parameters((True,), (False,))
    def test_save_load_pytree(self, use_async):
      self.save_and_wait(self.directory, self.pytree, use_async=use_async)
      loaded = self.load_and_wait(
          self.directory, self.abstract_pytree, use_async=use_async
      )
      test_utils.assert_tree_equal(self, self.pytree, loaded)

    @parameterized.parameters((True,), (False,))
    def test_load_default(self, use_async):
      self.save_and_wait(self.directory, self.pytree, use_async=use_async)
      loaded = self.load_and_wait(self.directory, use_async=use_async)
      test_utils.assert_tree_equal(self, self.pytree, loaded)

    def test_save_pytree_async(self):
      response = ocp.save_pytree_async(self.directory, self.pytree)
      self.assertFalse(self.directory.exists())  # Not finalized yet.
      # But a tmp dir should have been created.
      self.assertNotEmpty(list(self.directory.parent.iterdir()))
      response.result()
      self.assertTrue(self.directory.exists())
      loaded = ocp.load_pytree(self.directory)
      test_utils.assert_tree_equal(self, self.pytree, loaded)

    @parameterized.parameters(
        (tuple([]),),
        (dict(),),
        (list(),),
        (None,),
        (optax.EmptyState(),),
    )
    def test_empty_tree(self, tree):
      with self.assertRaisesRegex(ValueError, 'empty'):
        ocp.save_pytree(self.directory, tree)

    # Note the ommission of jax.Array, since this is covered in
    # several other tests.
    @parameterized.parameters(
        (np.arange(8),),
        (2,),
        (2.2,),
        ('foo',),
        (np.asarray(3.14),),
    )
    def test_standard_leaf_types(self, value):
      ocp.save_pytree(self.directory, dict(k=value))
      loaded = ocp.load_pytree(self.directory)
      if isinstance(value, np.ndarray):
        np.testing.assert_array_equal(loaded['k'], value)
      else:
        self.assertEqual(loaded['k'], value)

    def test_jax_array_leaf_types(self):
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',))
      # TODO(cpgaffney): Add support for missing arrays.
      values = {
          # 'simple_array': jnp.arange(16),
          # 'single_device_array': jnp.arange(8, device=jax.local_devices()[0]),
          'replicated_array': jnp.arange(
              12,
              device=jax.sharding.NamedSharding(
                  mesh, jax.sharding.PartitionSpec()
              ),
          ),
          'sharded_array': jnp.arange(
              32,
              device=jax.sharding.NamedSharding(
                  mesh, jax.sharding.PartitionSpec(('devices',))
              ),
          ),
          # 'single_device_cpu_array': jnp.arange(
          #     24, device=jax.local_devices(backend='cpu')[0]
          # ),
      }
      for k, v in values.items():
        with self.subTest(k):
          ocp.save_pytree(self.directory / k, [v])
          with self.subTest('with_abstract_pytree'):
            loaded = ocp.load_pytree(self.directory / k, [as_abstract_type(v)])
            test_utils.assert_tree_equal(self, [v], loaded)
          with self.subTest('without_abstract_pytree'):
            if multihost.is_pathways_backend():
              self.skipTest('Must provide abstract_pytree for Pathways.')
            loaded = ocp.load_pytree(self.directory / k)
            test_utils.assert_tree_equal(self, [v], loaded)

    def test_empty_array(self):
      value = np.ones(shape=(0,))
      with self.assertRaisesRegex(ValueError, 'zero size'):
        ocp.save_pytree(self.directory, dict(k=value))

    @parameterized.parameters(
        (complex(1.1, 2.2),),
        (2j,),
        # (b'foo',),  # TODO(cpgaffney): Unknown error in multihost.
        # (bytes(12),),  # TODO(cpgaffney): Unknown error in multihost.
    )
    def test_invalid_leaf_types(self, value):
      # TODO(cpgaffney): Consider improving the error message raised.
      with self.assertRaises(ValueError):
        ocp.save_pytree(self.directory, dict(k=value))

    # TODO(b/337105122): Add tests passing invalid abstract PyTrees.
    def test_flax_model(self):

      @flax.struct.dataclass
      class Params(flax.struct.PyTreeNode):
        params: PyTree
        opt_state: PyTree

      def make_state_with_optax():
        return Params(
            params=self.numpy_pytree,
            opt_state=(optax.EmptyState(), optax.EmptyState()),
        )

      def make_state_with_nones():
        return Params(
            params=self.numpy_pytree,
            opt_state=(None, None),
        )

      state = make_state_with_optax()
      ocp.save_pytree(self.directory, state)

      with self.subTest('with_abstract_state'):
        abstract_state = jax.tree.map(as_abstract_type, state)
        loaded = ocp.load_pytree(self.directory, abstract_state)
        test_utils.assert_tree_equal(self, state, loaded)

      with self.subTest('without_abstract_state'):
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        loaded = ocp.load_pytree(self.directory)
        expected_tree = tree_utils.serialize_tree(
            make_state_with_nones(),
            keep_empty_nodes=True,
        )
        test_utils.assert_tree_equal(self, expected_tree, loaded)

    @parameterized.parameters(
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x', 'y'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y', 'x'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x',))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y',))),
        (jax.sharding.PartitionSpec(('x', 'y')), jax.sharding.PartitionSpec()),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('x',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y', 'x')),
        ),
        (
            jax.sharding.PartitionSpec(('x',)),
            jax.sharding.PartitionSpec(('y',)),
        ),
    )
    def test_reshard(self, save_spec, load_spec):
      len_devices = len(jax.devices())
      self.assertGreaterEqual(len_devices, 4)
      mesh = jax.sharding.Mesh(
          np.asarray(jax.devices()).reshape((4, len_devices // 4)), ('x', 'y')
      )
      save_sharding = jax.sharding.NamedSharding(mesh, save_spec)
      tree = {'x': create_sharded_array(np.arange(len_devices), save_sharding)}
      ocp.save_pytree(self.directory, tree)

      load_sharding = jax.sharding.NamedSharding(mesh, load_spec)
      expected_tree = {
          'x': create_sharded_array(np.arange(len_devices), load_sharding)
      }
      abstract_tree = {'x': as_abstract_type(expected_tree['x'])}
      loaded = ocp.load_pytree(self.directory, abstract_pytree=abstract_tree)
      test_utils.assert_tree_equal(self, expected_tree, loaded)

    @parameterized.parameters(
        (np.int32, np.int16, np.int64),
        (np.int32, np.int64, np.int16),
        (np.int32, np.float32, np.int64),
        (np.int32, np.float32, np.float64),
        (np.float32, np.int16, np.int64),
        (np.float32, np.int64, np.int16),
        (np.float32, np.float32, np.int64),
        (np.float32, np.int32, np.int64),
        (np.int64, np.float32, jnp.bfloat16),
        (np.int64, jnp.bfloat16, np.int32),
        (jnp.bfloat16, np.int32, jnp.bfloat16),
    )
    def test_casting(self, original_dtype, save_dtype, load_dtype):
      sharding = jax.sharding.NamedSharding(
          jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',)),
          jax.sharding.PartitionSpec(),
      )
      tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=original_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=original_dtype),
      }
      save_casted_tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=save_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=save_dtype),
      }
      load_casted_tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=load_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=load_dtype),
      }

      create_array_storage_options_fn = (
          lambda k, v: ocp.options.ArrayStorageOptions(dtype=save_dtype)
      )
      with ocp.Context(
          pytree_options=ocp.options.PyTreeOptions(
              create_array_storage_options_fn=create_array_storage_options_fn
          )
      ):
        ocp.save_pytree(self.directory, tree)

      with self.subTest('with_abstract_tree'):
        abstract_tree = jax.tree.map(as_abstract_type, load_casted_tree)
        loaded = ocp.load_pytree(self.directory, abstract_tree)
        test_utils.assert_tree_equal(self, load_casted_tree, loaded)

      with self.subTest('without_abstract_tree'):
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        loaded = ocp.load_pytree(self.directory)
        test_utils.assert_tree_equal(self, save_casted_tree, loaded)

    # TODO(b/295313820): Improve mismatched-tree error messages.
    def test_mismatched_abstract_tree(self):
      ocp.save_pytree(self.directory, self.pytree)

      with self.subTest('subset_of_keys'):
        abstract_pytree = dict(self.abstract_pytree)
        del abstract_pytree['a'], abstract_pytree['b']
        with self.assertRaisesRegex(ValueError, 'Dict key mismatch'):
          ocp.load_pytree(self.directory, abstract_pytree)

      with self.subTest('superset_of_keys'):
        abstract_pytree = dict(self.abstract_pytree)
        abstract_pytree['z'] = as_abstract_type(np.arange(16))
        with self.assertRaisesRegex(ValueError, 'Dict key mismatch'):
          ocp.load_pytree(self.directory, abstract_pytree)

      with self.subTest('renamed_key'):
        abstract_pytree = dict(self.abstract_pytree)
        abstract_pytree['z'] = abstract_pytree.pop('a')
        with self.assertRaisesRegex(ValueError, 'Dict key mismatch'):
          ocp.load_pytree(self.directory, abstract_pytree)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorStore serialization helper functions."""

import math
import os
import re
from typing import Any, Dict, TypeAlias, Union

from absl import logging
from jax import numpy as jnp
from orbax.checkpoint._src.arrays import subchunking
from orbax.checkpoint._src.arrays import types
from orbax.checkpoint._src.metadata import array_metadata
import tensorstore as ts

JsonSpec: TypeAlias = dict[str, Any]
Shape: TypeAlias = types.Shape
DType: TypeAlias = types.DType
ArrayMetadata: TypeAlias = array_metadata.ArrayMetadata
ExtMetadata: TypeAlias = array_metadata.ExtMetadata

FILE_DRIVER = 'file'
DEFAULT_DRIVER = FILE_DRIVER

PROCESS_SUBDIR_PREFIX = 'ocdbt.process_'
_OCDBT_PROCESS_ID_RE = r'[A-Za-z0-9]+'
_DEFAULT_OCDBT_TARGET_DATA_FILE_SIZE = 2**31  # 2 GiB

ZARR_VER2 = 'zarr'
ZARR_VER3 = 'zarr3'

_GCS_PATH_RE = r'^gs://([^/]*)/(.*)$'

# Even if the data is equal to the fill value, we still want to write it
# to the checkpoint. This results in unnecessary writes in some edge
# cases, but it allows us to verify that data was actually written when
# later restoring.
# Must match `store_data_equal_to_fill_value` property in Orbax
# metadata.
STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE = True


_BASE_TS_CONTEXT = {
    'file_io_concurrency': {'limit': 128},
}
_DEFAULT_OCDBT_TS_CONTEXT = {
    **_BASE_TS_CONTEXT,
    # Provide cache pool for B-tree nodes to avoid repeated reads.
    # 100MB limit.
    **{'cache_pool#ocdbt': {'total_bytes_limit': 100000000}},
}

_REMOTE_URL_PREFIXES = ['gs://', 's3://']
_REMOTE_DRIVER_VALIDATIONS = [
    {'driver': 'gcs', 'path_regex': None},
    {'driver': 's3', 'path_regex': None},
]



def get_ts_context(*, use_ocdbt: bool = True) -> ts.Context:
  """Creates a TensorStore context object.

  For use with Orbax serialization APIs, or when directly opening a
  `TensorStore` object.

  Args:
    use_ocdbt: Whether to use OCDBT driver. Adds options specific to OCDBT if
      True.

  Returns:
    A TensorStore context object.
  """
  if use_ocdbt:
    return ts.Context(_DEFAULT_OCDBT_TS_CONTEXT)
  else:
    return ts.Context(_BASE_TS_CONTEXT)


### Building KvStore specs.


def _get_kvstore_for_gcs(ckpt_path: str) -> JsonSpec:
  m = re.fullmatch(_GCS_PATH_RE, ckpt_path, re.DOTALL)
  if m is None:
    raise ValueError(
        'The ckpt_path should contain the bucket name and the '
        f'file path inside the bucket. Got: {ckpt_path}'
    )
  gcs_bucket = m.group(1)
  path_without_bucket = m.group(2)
  return {'driver': 'gcs', 'bucket': gcs_bucket, 'path': path_without_bucket}


def build_kvstore_tspec(
    directory: str,
    name: str | None = None,
    *,
    use_ocdbt: bool = True,
    process_id: int | str | None = None,
) -> JsonSpec:
  """Constructs a spec for a Tensorstore KvStore.

  Args:
    directory: Base path (key prefix) of the KvStore, used by the underlying
      file driver.
    name: Name (filename) of the parameter.
    use_ocdbt: Whether to use OCDBT driver.
    process_id: [only used with OCDBT driver] If provided,
      `{directory}/ocdbt.process_{process_id}` path is used as the base path. If
      a string, must conform to [A-Za-z0-9]+ pattern.

  Returns:
    A Tensorstore KvStore spec in dictionary form.
  """
  default_driver = DEFAULT_DRIVER
  # Normalize path to exclude trailing '/'. In GCS path case, we will need to
  # fix the path prefix to add back the stripped '/'.
  directory = os.path.normpath(directory).replace('gs:/', 'gs://')
  is_gcs_path = directory.startswith('gs://')
  kv_spec = {}

  if use_ocdbt:
    if not is_gcs_path and not os.path.isabs(directory):
      raise ValueError(f'Checkpoint path should be absolute. Got {directory}')
    if process_id is not None:
      process_id = str(process_id)
      if re.fullmatch(_OCDBT_PROCESS_ID_RE, process_id) is None:
        raise ValueError(
            f'process_id must conform to {_OCDBT_PROCESS_ID_RE} pattern'
            f', got {process_id}'
        )
      directory = os.path.join(
          directory, f'{PROCESS_SUBDIR_PREFIX}{process_id}'
      )
    base_driver_spec = (
        directory
        if is_gcs_path
        else {'driver': default_driver, 'path': str(directory)}
    )
    kv_spec.update({
        'driver': 'ocdbt',
        'base': base_driver_spec,
    })
    if name is not None:
      kv_spec['path'] = name

    kv_spec.update({  # pytype: disable=attribute-error
        # References the cache specified in ts.Context.
        'cache_pool': 'cache_pool#ocdbt',
    })

    if is_remote_storage(kv_spec):
      kv_spec.update({  # pytype: disable=attribute-error
          # Enable read coalescing.  This feature merges adjacent read_ops into
          # one, which could reduce I/O ops by a factor of 10. This is
          # especially beneficial for unstacked models.
          'experimental_read_coalescing_threshold_bytes': 1000000,
          'experimental_read_coalescing_merged_bytes': 500000000000,
          'experimental_read_coalescing_interval': '1ms',
      })
  else:
    if name is None:
      path = directory
    else:
      path = os.path.join(directory, name)
    if is_gcs_path:
      kv_spec = _get_kvstore_for_gcs(path)
    else:
      kv_spec = {'driver': default_driver, 'path': path}

  return kv_spec


def add_ocdbt_write_options(
    kvstore_tspec: JsonSpec,
    target_data_file_size: int | None = None,
) -> None:
  """Adds write-specific options to a TensorStore OCDBT KVStore spec."""
  if target_data_file_size is not None:
    # TODO: b/354139177 - disallow too small values, too.
    if target_data_file_size < 0:
      raise ValueError(
          'OCDBT target_data_file_size must be >= 0, where 0 means no limit'
          f'; got {target_data_file_size}'
      )
    kvstore_tspec['target_data_file_size'] = target_data_file_size

  kvstore_tspec['config'] = {
      # Store .zarray metadata inline but not large chunks.
      'max_inline_value_bytes': 1024,
      # Large value allows a single root node to support faster traversal.
      'max_decoded_node_bytes': 100000000,
      # There won't be any concurrent writes by multiple machines to the same
      # OCDBT database.  Therefore, we can use the simpler and more efficient
      # single-file manifest format in all cases.
      'manifest_kind': 'single',
  }
  # assume_config avoids writing an initial empty manifest to ensure a
  # consistent configuration, since Orbax never writes to the same OCDBT
  # database concurrently from multiple processes.
  kvstore_tspec.update(assume_config=True)


### Building Zarr array metadata.


def build_zarr_shard_and_chunk_metadata(
    *,
    global_shape: Shape,
    shard_shape: Shape,
    use_zarr3: bool,
    chunk_shape: Shape,
) -> JsonSpec:
  """Constructs Zarr metadata for TensorStore array write spec."""
  metadata = {'shape': global_shape}

  if not use_zarr3:
    # Zarr v2.
    metadata['chunks'] = chunk_shape
    metadata['compressor'] = {'id': 'zstd'}
  else:
    # Zarr v3.
    metadata['chunk_grid'] = {
        'name': 'regular',
        'configuration': {
            'chunk_shape': chunk_shape,
        },
    }
    # TODO: b/354139177 - Consider if using write shape equal to shard shape and
    # read shape equal to chosen chunk shape would be a better setting.
    del shard_shape  # Currently unused.
    metadata['codecs'] = [
        {
            'name': 'sharding_indexed',
            'configuration': {
                'chunk_shape': chunk_shape,
                'codecs': [
                    {'name': 'bytes', 'configuration': {'endian': 'little'}},
                    {'name': 'zstd'},
                ],
                'index_codecs': [
                    {'name': 'bytes', 'configuration': {'endian': 'little'}},
                    {'name': 'crc32c'},
                ],
                'index_location': 'end',
            },
        },
    ]

  return metadata


def calculate_chunk_byte_size(
    write_shape: Shape,
    dtype: DType,
    *,
    chunk_byte_size: int | None,
    ocdbt_target_data_file_size: int | None = None,
) -> int | None:
  """Selects chunk byte size to fit both target data file and chunk sizes."""
  # Check if the chunk size would exceed ocdbt target file size.
  if ocdbt_target_data_file_size is None:
    # Set to default used by TensorStore
    # (from https://google.github.io/tensorstore/kvstore/ocdbt/index.html).
    ocdbt_target_data_file_size = _DEFAULT_OCDBT_TARGET_DATA_FILE_SIZE

  if ocdbt_target_data_file_size == 0:
    # No limit.
    return chunk_byte_size

  if chunk_byte_size is None:
    write_nbytes = math.prod(write_shape) * dtype.itemsize
    if write_nbytes > ocdbt_target_data_file_size:
      chunk_byte_size = ocdbt_target_data_file_size
    else:
      # Let chunk_byte_size stay None.
      chunk_byte_size = None
  else:
    chunk_byte_size = min(chunk_byte_size, ocdbt_target_data_file_size)
  return chunk_byte_size


### Building TensorStore array specs.


def _maybe_add_cast_to_write_spec(
    array_tspec: JsonSpec,
    *,
    dtype: DType,
    target_dtype: DType,
) -> JsonSpec:
  """Adds cast driver to a write array TensorStore spec, if needed."""
  if target_dtype == dtype:
    array_tspec['dtype'] = jnp.dtype(dtype).name
    return array_tspec

  array_tspec = {
      'base': array_tspec,
      'driver': 'cast',
  }
  # Origin dtype.
  array_tspec['dtype'] = jnp.dtype(dtype).name
  # Destination dtype.
  array_tspec['base']['dtype'] = jnp.dtype(target_dtype).name
  return array_tspec


class ArrayWriteSpec:
  """Full TensorStore spec for writing an array."""

  def __init__(
      self,
      directory: str,
      relative_array_filename: str,
      *,
      global_shape: Shape,
      write_shape: Shape,
      dtype: DType,
      target_dtype: DType | None = None,
      chunk_byte_size: int | None = None,
      shard_axes: tuple[int, ...] = (),
      use_zarr3: bool = False,
      use_ocdbt: bool,
      ocdbt_target_data_file_size: int | None = None,
      process_id: int | str | None = None,
      metadata_key: str | None = None,
      ext_metadata: ExtMetadata | None = None,
  ):
    """Builds a TensorStore spec for writing an array."""
    # Construct the underlying KvStore spec.
    kvstore_tspec = build_kvstore_tspec(
        directory,
        name=relative_array_filename,
        use_ocdbt=use_ocdbt,
        process_id=process_id,
    )
    # Construct the top-level array spec.
    tspec = {
        'driver': ZARR_VER3 if use_zarr3 else ZARR_VER2,
        'kvstore': kvstore_tspec,
        'recheck_cached_data': False,
        'recheck_cached_metadata': False,
        'store_data_equal_to_fill_value': STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE,
    }
    if metadata_key is not None:
      tspec['metadata_key'] = metadata_key

    target_storage_dtype = target_dtype or dtype

    # Choose target file and chunk byte sizes.
    if use_ocdbt:
      add_ocdbt_write_options(
          tspec['kvstore'],
          ocdbt_target_data_file_size,
      )
      chunk_byte_size = calculate_chunk_byte_size(
          write_shape,
          target_storage_dtype,
          chunk_byte_size=chunk_byte_size,
          ocdbt_target_data_file_size=ocdbt_target_data_file_size,
      )
    # Choose chunk shape.
    chunk_shape = subchunking.choose_chunk_shape(
        global_shape,
        write_shape,
        target_storage_dtype,
        chunk_byte_size,
        shard_axes=shard_axes,
    )
    if chunk_shape != write_shape:
      logging.info(
          'Array name: %r, global shape: %r, write shape: %r, chosen chunk'
          ' shape: %r',
          relative_array_filename,
          global_shape,
          write_shape,
          chunk_shape,
      )
    # Construct Zarr chunk metadata.
    tspec['metadata'] = build_zarr_shard_and_chunk_metadata(
        global_shape=global_shape,
        shard_shape=write_shape,
        use_zarr3=use_zarr3,
        chunk_shape=chunk_shape,
    )

    # Keep the metadata in a separate field.
    self._metadata = ArrayMetadata(
        param_name=relative_array_filename,
        shape=global_shape,
        dtype=target_storage_dtype,
        write_shape=write_shape,
        chunk_shape=chunk_shape,
        use_ocdbt=use_ocdbt,
        use_zarr3=use_zarr3,
        ext_metadata=ext_metadata,
    )
    # Wrap spec into `cast` driver if needed, and keep it in a separate field.
    self._json_spec = _maybe_add_cast_to_write_spec(
        tspec,
        dtype=dtype,
        target_dtype=target_storage_dtype,
    )

  @property
  def json(self) -> JsonSpec:
    """Spec to be used to open a TensorStore for writing the array."""
    return self._json_spec

  @property
  def metadata(self) -> ArrayMetadata:
    """Checkpoint-relevant TensorStore metadata of the array."""
    return self._metadata


def is_remote_storage(tspec: Union[Dict[str, Any], str]) -> bool:
  """Detect if user is using remote storages.

  This can detect common defines and unable to detect some corner cases such as
  using gcsfuse.

  Args:
    tspec: Tensorstore spec.

  Returns:
    True if the spec is using remote storage.
  """
  if isinstance(tspec, str):
    # KvStoreUrl
    if re.match(rf'^({"|".join(_REMOTE_URL_PREFIXES)})', tspec):
      return True
    else:
      return False

  for key in ('base', 'kvstore'):
    if key in tspec:
      return is_remote_storage(tspec[key])

  if 'driver' in tspec:
    for rule in _REMOTE_DRIVER_VALIDATIONS:
      if tspec['driver'] == rule['driver']:
        if rule['path_regex'] is None:
          return True

        # check if path matches the regex.
        if re.match(rule['path_regex'], tspec['path']):
          return True

  return False



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for loading."""

import orbax.checkpoint as ocp
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


def _get_concurrent_gb(concurrent_bytes: int | None) -> int | None:
  if concurrent_bytes:
    return max(int(concurrent_bytes / 1e9), 1)
  return None


def load_pytree(
    directory: path_types.PathLike,
    abstract_pytree: (
        tree_types.PyTreeOf[tree_types.AbstractLeafType] | None
    ) = None,
) -> tree_types.PyTreeOf[tree_types.LeafType]:
  """Loads a PyTree.

  The operation blocks until complete. For improved performance, consider using
  `load_async` instead.

  Args:
    directory: The directory to load the checkpoint from.
    abstract_pytree: Provides a tree structure for the checkpoint to be restored
      into. May be omitted to load exactly as saved, but this is much more
      brittle than providing the tree. Providing the `abstract_tree` guarantees
      two things: (1) The restored tree will exactly match the structure of
      `abstract_pytree` (or raise an error if it is impossible to guarantee
      this). For example, if `abstract_pytree` is a custom object registered as
      a PyTree, the checkpoint will be restored as the same object, if possible.
      (2) The leaves of the restored tree will be restored with the properties
      indicated by the abstract leaves. For example, if a leaf in
      `abstract_pytree` is a `jax.ShapeDtypeStruct`, the restored leaf will be a
      `jax.Array` with the same shape and dtype. Each `AbstractLeafType` has a

  Returns:
    The restored PyTree.
  """
  context = context_lib.get_context()

  if context.pytree_options.partial_load:
    raise NotImplementedError('Partial loading is not yet supported.')


  handler_registry = ocp.handlers.create_default_handler_registry(
      pytree=ocp.PyTreeCheckpointHandler(
          restore_concurrent_gb=_get_concurrent_gb(
              context.pytree_options.restore_concurrent_bytes
          ),
      )
  )
  ckptr = ocp.Checkpointer(
      ocp.CompositeCheckpointHandler(handler_registry=handler_registry)
  )
  args = ocp.args.Composite(
      pytree=ocp.args.PyTreeRestore(
          abstract_pytree,
          restore_args=ocp.checkpoint_utils.construct_restore_args(
              abstract_pytree
          ),
      )
  )
  return ckptr.restore(directory, args=args).pytree


def load_pytree_async(
    directory: path_types.PathLike,
    abstract_pytree: (
        tree_types.PyTreeOf[tree_types.AbstractLeafType] | None
    ) = None,
) -> async_types.AsyncResponse[tree_types.PyTreeOf[tree_types.LeafType]]:
  """Loads a PyTree asynchronously. Not yet implemented."""
  del directory, abstract_pytree
  raise NotImplementedError('Asynchronous loading is not yet supported.')



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines types for `CheckpointableHandler`."""

from typing import Awaitable, Protocol, TypeVar
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types


T = TypeVar('T')
AbstractT = TypeVar('AbstractT')
MetadataT = TypeVar('MetadataT')

PathLike = path_types.PathLike
AsyncResponse = async_types.AsyncResponse


class CheckpointableHandler(Protocol[T, AbstractT, MetadataT]):
  """An interface that defines save/load logic for a `checkpointable` object.

  NOTE: Prefer to use `Checkpointable` interface when possible.

  A "checkpointable" is a fundamental concept in Orbax. A “checkpointable”
  refers to a logical piece of the checkpoint that is distinct in some way from
  other pieces. Checkpointables are separable; they may or may not be loaded
  concurrently and some may be omitted from the checkpoint entirely.
  Checkpointables are often represented by different types, and have different
  representations on disk. The quintessential example is model params vs.
  dataset.

  A PyTree of arrays, representing model parameters, is the most basic
  "checkpointable". A singular array is also a checkpointable.

  In most contexts, when dealing with just a PyTree, the API of choice is::

    ocp.save_pytree(directory, pytree)

  The concept of "checkpointable" is not so obvious in this case. When dealing
  with multiple objects, we can use::

    ocp.save_checkpointables(
        directory,
        dict(
            pytree=model_params,
            dataset=dataset_iterator,
            # other checkpointables, e.g. extra metadata, etc.
        ),
    )

  Now, it is easy to simply skip loading the dataset, as is commonly desired
  when running evals or inference::

    ocp.load_checkpointables(
        directory,
        dict(
            pytree=abstract_model_params,
        ),
    )
    # Equivalently,
    ocp.load_pytree(directory, abstract_model_params)

  With the methods defined in this Protocol (`save`, `load`),
  logic within the method itself is executed in the main thread,
  in a blocking fashion. Additional logic can be executed in the background by
  returning an `Awaitable` function (which itself may return a result).

  TODO(b/398249409) Include more details on implementing this Protocol.
  """

  async def save(
      self, directory: path_types.PathLike, checkpointable: T
  ) -> Awaitable[None]:
    """Saves the given `checkpointable` to the given `directory`.

    Save should perform any operations that need to block the main thread, such
    as device-to-host copying of on-device arrays. It then creates a background
    operation to continue writing the object to the storage location.

    Args:
      directory: The directory to save the checkpoint to.
      checkpointable: The checkpointable object to save.

    Returns:
      An `Awaitable`. This object represents the result of the save
      operation running in the background.
    """
    ...

  async def load(
      self,
      directory: path_types.PathLike,
      abstract_checkpointable: AbstractT | None = None,
  ) -> Awaitable[T]:
    """Loads the checkpointable from the given `directory`.

    Args:
      directory: The directory to load the checkpoint from.
      abstract_checkpointable: An optional abstract representation of the
        checkpointable to load. If provided, this is used to provide properties
        to guide the restoration logic of the checkpoint. In the case of arrays,
        for example, this conveys properties like shape and dtype, for casting
        and reshaping. In some cases, no information is needed, and `AbstractT`
        may always be None. In other cases, the abstract representation may be a
        hard requirement for loading.

    Returns:
      An `Awaitable` that continues to load the checkpointable in the background
      and returns the loaded checkpointable when complete.
    """
    ...

  async def metadata(
      self, directory: path_types.PathLike
  ) -> MetadataT:
    """Returns the metadata for the given `directory`.

    The logic in this method must be executed fully in the main thread; metadata
    access is expected to be cheap and fast.

    Args:
      directory: The directory where the checkpoint is located.

    Returns:
      MetadataT: The metadata is an `MetadataT`,
      which is the abstract representation of
      the checkpointable. `MetadataT` differs from `AbstractT` in that it may
      contain additional properties that cannot be directly consumed to
      customize loading behavior, but are nevertheless present and useful
      to know about in some cases.
    """
    ...



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines implementations of AsyncResponse."""

from typing import Coroutine, Sequence, TypeVar

from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import types


AsyncResponse = types.AsyncResponse

T = TypeVar('T')


class _DefaultAsyncResponse(AsyncResponse[T]):
  """Represents the result of a background function call.

  May send signals to indicate that the commit has completed. Can also receive
  signals to indicate that the commit should proceed.
  """

  def __init__(
      self,
      coroutine: Coroutine[None, None, T],
      *,
      name: str | None = None,
      send_signals: (
          Sequence[synchronization.HandlerAwaitableSignal] | None
      ) = None,
      receive_signals: (
          Sequence[synchronization.HandlerAwaitableSignal] | None
      ) = None,
      timeout_secs: int = 600,
      operation_id: str | None = None,
  ):
    """Constructor.

    Args:
      coroutine: An coroutine (defined with e.g. `async def foo(a, b)`). The
        awaitable function is passed as `foo(a, b)`, without awaiting, since
        this will be done in the background thread.
      name: The name of the thread.
      send_signals: Signals to send to indicate that the commit has completed.
      receive_signals: Signals to wait for before proceeding with the commit.
      timeout_secs: Timeout in seconds for waiting for signals.
      operation_id: The operation id to use for the barrier keys. If None, the
        current operation id is used.
    """
    send_signals = send_signals or []
    receive_signals = receive_signals or []
    self._t = future._SignalingThread(
        send_signals=send_signals,
        receive_signals=receive_signals,
        timeout_secs=timeout_secs,
        operation_id=operation_id,
        target=lambda: asyncio_utils.run_sync(coroutine),
        name=name,
    )
    self._t.start()

  def result(self, timeout: float | None = None) -> T:
    """Waits for the commit to complete."""
    return self._t.result(timeout=timeout)


class _AsyncResponseAwaitingContractedSignals(AsyncResponse[T]):
  """Represents the result of a background function call.

  May send signals to indicate that the function has completed. Waits for all
  awaitable signals in the `AWAITABLE_SIGNALS_CONTRACT` to be set before
  proceeding with the commit.
  """

  def __init__(
      self,
      coroutine: Coroutine[None, None, T],
      *,
      name: str | None = None,
      send_signals: (
          Sequence[synchronization.HandlerAwaitableSignal] | None
      ) = None,
      timeout_secs: int = 600,
      operation_id: str | None = None,
  ):
    """Constructor.

    Synchronously gets all awaitable signals in the contract and waits to
    receive them in background before proceeding with the commit if JAX
    distributed client is initialized.

    Args:
      coroutine: An awaitable (defined with e.g. `async def foo(a, b)`). The
        awaitable function is passed as `foo(a, b)`, without awaiting, since
        this will be done in the background thread.
      name: The name of the thread.
      send_signals: Signals to send to indicate that the commit has completed.
      timeout_secs: Timeout in seconds for waiting for signals.
      operation_id: The operation id to use for the barrier keys. If None, the
        current operation id is used.
    """
    receive_signals = []
    if multihost.is_jax_distributed_client_initialized():
      receive_signals = future.get_awaitable_signals_from_contract()
    self._response = _DefaultAsyncResponse[T](
        coroutine,
        name=name,
        send_signals=send_signals,
        receive_signals=receive_signals,
        timeout_secs=timeout_secs,
        operation_id=operation_id,
    )

  def result(self, timeout: float | None = None) -> T:
    return self._response.result(timeout=timeout)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""High-level checkpoint utils provided for user convenience."""

import contextlib
import time
from typing import Any, Callable, Iterator, Optional

from absl import logging
from etils import epath
import jax
from jax.experimental import layout
import numpy as np
from orbax.checkpoint import utils
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import step as step_lib
from orbax.checkpoint._src.path.snapshot import snapshot as snapshot_lib
from orbax.checkpoint._src.serialization import type_handlers


PyTree = Any
STANDARD_ARRAY_TYPES = (int, float, np.ndarray, jax.Array)
_SNAPSHOTS = '_SNAPSHOTS'
Layout = layout.Layout


def _init_step_name_format(
    step_name_format: Optional[step_lib.NameFormat[step_lib.Metadata]] = None,
    step_prefix: Optional[str] = None,
    step_format_fixed_length: Optional[int] = None,
):
  return step_name_format or step_lib.standard_name_format(
      step_prefix=step_prefix,
      step_format_fixed_length=step_format_fixed_length,
  )


def get_snapshot_dir_from_step_dir(
    step_dir: epath.Path, snapshot_dir: Optional[epath.Path] = None
) -> epath.Path:
  """Returns the snapshot directory from the step directory."""
  if snapshot_dir is None:
    snapshot_dir = step_dir.parent / _SNAPSHOTS
  new_path = snapshot_dir / step_dir.name
  return new_path


def _snapshot_checkpoint(
    checkpoint_dir: epath.Path,
    step: int,
    step_name_format: step_lib.NameFormat[step_lib.Metadata],
    snapshot_dir: Optional[epath.Path] = None,
):
  """Uses `Snapshot` class to create a cheap "copy" of the checkpoint."""
  if multihost.process_index() == 0:
    logging.info('Snpashotting step: %d.', step)
    step_dir = step_name_format.find_step(checkpoint_dir, step).path
    if not step_dir.exists():
      raise ValueError(f'Step directory {step_dir} does not exist.')

    if snapshot_dir is None:
      snapshot_dir = checkpoint_dir / _SNAPSHOTS
    if not snapshot_dir.exists():
      try:
        snapshot_dir.mkdir(parents=True, exist_ok=True)
      except OSError as e:
        raise ValueError(
            f'Failed to create snapshot directory {snapshot_dir}. Please'
            ' provide a snapshot directory instead.'
        ) from e

    snapshot_path = get_snapshot_dir_from_step_dir(step_dir, snapshot_dir)
    if epath.Path(snapshot_path).exists():
      return True
    snapshot_impl = snapshot_lib.create_instance(str(snapshot_path))
    dst_path = snapshot_impl.create_snapshot(str(step_dir), str(snapshot_path))
    if str(snapshot_path) == dst_path:
      return True
  return False


def _release_snapshot(
    checkpoint_dir: epath.Path,
    step: int,
    step_name_format: step_lib.NameFormat[step_lib.Metadata],
    snapshot_dir: Optional[epath.Path] = None,
):
  """Releases snapshot by deleting the snapshot of the checkpoint."""
  if multihost.process_index() == 0:
    logging.info('Releasing snapshot at step: %d.', step)
    if snapshot_dir is None:
      snapshot_dir = checkpoint_dir / _SNAPSHOTS
    snapshot_path = snapshot_dir / step_name_format.build_name(step)
    snapshot_impl = snapshot_lib.create_instance(str(snapshot_path))
    snapshot_impl.release_snapshot(str(snapshot_path))


def _reached_desired_step(step: int, until_step: Optional[int]) -> bool:
  if step is None:
    return False
  elif until_step is None:
    return True
  elif step >= until_step:
    return True
  return False


def _wait_for_new_checkpoint(
    checkpoint_dir: epath.Path,
    *,
    step_name_format: step_lib.NameFormat[step_lib.Metadata],
    until_step: Optional[int] = None,
    seconds_to_sleep: int = 1,
    timeout: Optional[int] = None,
    timeout_fn: Optional[Callable[[], bool]] = None,
    snapshot_dir: Optional[epath.Path] = None,
) -> int:
  """See documentation for wait_for_new_checkpoint."""
  start = time.time()
  stop_time = start + timeout if timeout is not None else None

  def _sleep_and_maybe_exit():
    if stop_time is not None and time.time() + seconds_to_sleep > stop_time:
      if timeout_fn is None:
        return True
      elif timeout_fn():  # Only exit when timeout_fn indicates completion.
        return True
    logging.info('Sleeping for %d seconds.', seconds_to_sleep)
    time.sleep(seconds_to_sleep)
    return False

  log_str = f'Waiting for new checkpoint at {checkpoint_dir}. '
  if until_step is not None:
    log_str += f'Waiting until step {until_step} is reached. '
  if timeout is not None:
    log_str += f'Will time out after {timeout} seconds. '
  logging.info(log_str)

  result = -1
  if multihost.process_index() == 0:
    while True:
      if not checkpoint_dir.exists():
        if _sleep_and_maybe_exit():
          break
        continue  # continue waiting until directory creation or timeout.

      steps = utils.checkpoint_steps(checkpoint_dir)
      checkpoint_step = max(steps) if steps else None
      if _reached_desired_step(checkpoint_step, until_step):
        if not _snapshot_checkpoint(
            checkpoint_dir, checkpoint_step, step_name_format, snapshot_dir
        ):
          continue
        result = checkpoint_step
        break
      elif _sleep_and_maybe_exit():
        break

  result = multihost.broadcast_one_to_all(np.int32(result)).item()
  wait_duration = time.time() - start
  jax.monitoring.record_event_duration_secs(
      '/jax/orbax/checkpoint_utils/wait_duration', wait_duration
  )

  if result == -1:
    logging.info('Timed out waiting for new checkpoint. Returning -1.')
  else:
    logging.info('Found new checkpoint step: %d.', result)
  return result


@contextlib.contextmanager
def wait_for_new_checkpoint(
    checkpoint_dir: epath.Path,
    *,
    until_step: Optional[int] = None,
    seconds_to_sleep: int = 1,
    timeout: Optional[int] = None,
    timeout_fn: Optional[Callable[[], bool]] = None,
    step_prefix: Optional[str] = None,
    step_format_fixed_length: Optional[int] = None,
    step_name_format: Optional[step_lib.NameFormat[step_lib.Metadata]] = None,
    snapshot_dir: Optional[epath.Path] = None,
):
  """Waits until a new checkpoint file is found.

  Automatically snapshots any checkpoint that is returned, and releases the
  snapshot of the checkpoint when execution returns to this function.

  Args:
    checkpoint_dir: The directory in which checkpoints are saved.
    until_step: If specified, waits until a step greater than or equal to
      `until_step` has been found. If set to None (default), returns the first
      step found.
    seconds_to_sleep: The number of seconds to sleep for before looking for a
      new checkpoint.
    timeout: The maximum number of seconds to wait. If left as `None`, then the
      process will wait indefinitely.
    timeout_fn: Optional function to call after a timeout.  If the function
      returns True, then it means that no new checkpoints will be generated and
      the iterator will exit.  The function is called with no arguments.
    step_prefix: A prefix applied to step numbers (e.g. <prefix>_42).
    step_format_fixed_length: Expects to find checkpoint step directories with
      exactly this number of digits (leading zeros if necessary).
    step_name_format: Step NameFormat used to find step under given root
      directory. If provided, `step_prefix` and `step_format_fixed_length` are
      ignored.
    snapshot_dir: The directory in which snapshots are saved. If not provided,
      the snapshot directory is created as `checkpoint_dir / _SNAPSHOTS`.

  Yields:
    a new checkpoint step, or -1 if the timeout was reached.
  """
  step_name_format = _init_step_name_format(
      step_name_format=step_name_format,
      step_prefix=step_prefix,
      step_format_fixed_length=step_format_fixed_length,
  )
  step = _wait_for_new_checkpoint(
      checkpoint_dir,
      step_name_format=step_name_format,
      until_step=until_step,
      seconds_to_sleep=seconds_to_sleep,
      timeout=timeout,
      timeout_fn=timeout_fn,
      snapshot_dir=snapshot_dir,
  )
  try:
    yield step
  finally:
    # Release snapshot on the checkpoint step.
    if step != -1:
      logging.info(
          'Releasing snapshot for step: %d after releasing control.', step
      )
      _release_snapshot(checkpoint_dir, step, step_name_format, snapshot_dir)


def checkpoints_iterator(
    checkpoint_dir: epath.PathLike,
    *,
    min_interval_secs: int = 0,
    seconds_to_sleep: int = 1,
    timeout: Optional[int] = None,
    timeout_fn: Optional[Callable[[], bool]] = None,
    step_prefix: Optional[str] = None,
    step_format_fixed_length: Optional[int] = None,
    step_name_format: Optional[step_lib.NameFormat[step_lib.Metadata]] = None,
    snapshot_dir: Optional[epath.Path] = None,
) -> Iterator[int]:
  """Continuously yield new checkpoint files as they appear.

  Based on the equivalent TF method.

  The iterator only checks for new checkpoints when control flow has been
  reverted to it. This means it can miss checkpoints if your code takes longer
  to run between iterations than `min_interval_secs` or the interval at which
  new checkpoints are written.

  Warning: If CheckpointManager is running in a different process for training
  and is cleaning up old checkpoints (via the `max_to_keep` argument), steps
  returned by this function may not be valid after being clean up by another
  process. In this case, `max_to_keep` should be increased (suggested value: 5)

  The `timeout` argument is the maximum number of seconds to block waiting for
  a new checkpoint.  It is used in combination with the `timeout_fn` as
  follows:

  * If the timeout expires and no `timeout_fn` was specified, the iterator
    stops yielding.
  * If a `timeout_fn` was specified, that function is called and if it returns
    a true boolean value the iterator stops yielding.
  * If the function returns a false boolean value then the iterator resumes the
    wait for new checkpoints.  At this point the timeout logic applies again.

  This behavior gives control to callers on what to do if checkpoints do not
  come fast enough or stop being generated.  For example, if callers have a way
  to detect that the training has stopped and know that no new checkpoints
  will be generated, they can provide a `timeout_fn` that returns `True` when
  the training has stopped.  If they know that the training is still going on
  they return `False` instead.

  Args:
    checkpoint_dir: The directory in which checkpoints are saved.
    min_interval_secs: The minimum number of seconds between yielding
      checkpoints.
    seconds_to_sleep: Seconds to sleep if a checkpoint is not found. Note the
      difference with min_interval_secs, which puts a lower bound on how when a
      new checkpoint will be looked for after yielding one checkpoint.
      seconds_to_sleep instead specifies how we should sleep for if no new
      checkpoints are found. Note that the timeout is only checked when not
      sleeping, so a `seconds_to_sleep` longer than the timeout would result in
      timing out after `seconds_to_sleep` seconds rather than `timeout` seconds.
    timeout: The maximum number of seconds to wait between checkpoints. The
      function will time out if `timeout` seconds have passed since a new
      checkpoint step was found. If left as `None`, then the process will wait
      indefinitely.
    timeout_fn: Optional function called after a timeout. If the function
      returns True, then it means that no new checkpoints will be generated and
      the iterator will exit. The function is called with no arguments.
    step_prefix: A prefix applied to step numbers (e.g. <prefix>_42).
    step_format_fixed_length: Expects to find checkpoint step directories with
      exactly this number of digits (leading zeros if necessary).
    step_name_format: Step NameFormat used to find step under given root
      directory. If provided, `step_prefix` and `step_format_fixed_length` are
      ignored.
    snapshot_dir: The directory in which snapshots are saved. If not provided,
      the snapshot directory is created as `checkpoint_dir / _SNAPSHOTS`.

  Yields:
    Integer step numbers of the latest checkpoints as they arrive.
  """
  checkpoint_dir = epath.Path(checkpoint_dir)
  step_name_format = _init_step_name_format(
      step_name_format=step_name_format,
      step_prefix=step_prefix,
      step_format_fixed_length=step_format_fixed_length,
  )
  snapshot_impl = snapshot_lib.create_instance(str(checkpoint_dir))
  if snapshot_dir is None:
    snapshot_dir = checkpoint_dir / _SNAPSHOTS
  if snapshot_dir.exists():
    for step_dir in snapshot_dir.iterdir():
      snapshot_impl.release_snapshot(str(step_dir))
  checkpoint_step = None
  while True:
    until_step = checkpoint_step + 1 if checkpoint_step is not None else None
    with wait_for_new_checkpoint(
        checkpoint_dir,
        until_step=until_step,
        seconds_to_sleep=seconds_to_sleep,
        timeout=timeout,
        timeout_fn=timeout_fn,
        step_name_format=step_name_format,
        snapshot_dir=snapshot_dir,
    ) as new_checkpoint_step:
      if new_checkpoint_step == -1:
        if not timeout_fn:
          logging.info('Timed-out waiting for a checkpoint.')
          return
        if timeout_fn():
          # The timeout_fn indicated that we are truly done.
          return
        else:
          # The timeout_fn indicated that more checkpoints may come.
          continue
      start = time.time()
      checkpoint_step = new_checkpoint_step
      yield checkpoint_step
    time_to_next_eval = start + min_interval_secs - time.time()
    if time_to_next_eval > 0:
      time.sleep(time_to_next_eval)


def _python_type_from_dtype(dtype):
  if hasattr(dtype, 'type'):
    return type(dtype.type(0).item())
  else:
    return type(dtype(0).item())


def construct_restore_args(
    target: PyTree,
    sharding_tree: Optional[PyTree] = None,
    set_global_shape: bool = True,
    support_layout: bool = False,
) -> PyTree:
  """Creates restore_args given a target PyTree.

  This method should be used in conjunction with a CheckpointManager or
  Checkpointer that wraps a PyTreeCheckpointHandler.

  For example::

    mngr = CheckpointManager(path, Checkpointer(PyTreeCheckpointHandler()))
    restore_args = construct_restore_args(train_state, train_state_sharding)
    restore_kwargs = {'restore_args': restore_args}
    mngr.restore(..., restore_kwargs=restore_kwargs)

  OR::

    mngr = CheckpointManager(path, {
        'train_state': Checkpointer(PyTreeCheckpointHandler())
    })
    restore_args = construct_restore_args(train_state, train_state_sharding)
    restore_kwargs = {'train_state': {'restore_args': restore_args} }
    mngr.restore(..., restore_kwargs=restore_kwargs)

  OR::

    ckptr = Checkpointer(PyTreeCheckpointHandler())
    restore_args = construct_restore_args(train_state, train_state_sharding)
    ckptr.restore(..., restore_args=restore_args)

  If a leaf in target is a np.ndarray, or int, or string, for example, a
  corresponding value for that leaf must be provided in axes_tree, but will be
  ignored.

  Args:
    target: The returned PyTree will match the structure of `target`. `target`
      may contain `value_metadata.Metadata`, real scalar or array values, or may
      contain jax.ShapeDtypeStruct.
    sharding_tree: A PyTree matching `target` which will be used to set the
      restoration sharding. If not provided, sharding will default to the
      shardings specified by `target`.
    set_global_shape: If true, set the `global_shape` field of ArrayRestoreArgs.
    support_layout: If true, layout is extracted from jax.Array or
      jax.ShapeDtypeStruct.

  Returns:
    A PyTree matching target of RestoreArgs (or ArrayRestoreArgs) objects.
  """

  def _array_restore_args(
      value: Any,
      sharding: Optional[jax.sharding.Sharding | Layout],
      dtype: Optional[np.dtype] = None,
  ) -> type_handlers.ArrayRestoreArgs:
    return type_handlers.ArrayRestoreArgs(
        restore_type=jax.Array,
        sharding=sharding,
        global_shape=value.shape if set_global_shape else None,
        dtype=dtype,
    )

  def _restore_args(
      value: Any,
      sharding: Optional[jax.sharding.Sharding],
  ) -> type_handlers.RestoreArgs:
    if isinstance(value, jax.ShapeDtypeStruct):
      if sharding is None:
        return type_handlers.RestoreArgs(dtype=value.dtype)
      else:
        return _array_restore_args(value, sharding, value.dtype)
    elif isinstance(value, value_metadata.Metadata):
      if isinstance(value, value_metadata.StringMetadata):
        return type_handlers.RestoreArgs(restore_type=str)
      elif isinstance(value, value_metadata.ScalarMetadata):
        return type_handlers.RestoreArgs(
            restore_type=_python_type_from_dtype(value.dtype), dtype=value.dtype
        )
      elif isinstance(value, value_metadata.ArrayMetadata):
        if sharding is None:
          return type_handlers.RestoreArgs(
              restore_type=np.ndarray, dtype=value.dtype
          )
        else:
          return _array_restore_args(value, sharding, value.dtype)
      else:
        raise ValueError(f'Unsupported value_metadata class: {type(value)}.')
    elif isinstance(value, STANDARD_ARRAY_TYPES):
      if isinstance(value, np.ndarray):
        return type_handlers.RestoreArgs(
            restore_type=type(value), dtype=value.dtype
        )
      elif isinstance(value, jax.Array):
        return _array_restore_args(value, sharding, value.dtype)
      else:
        return type_handlers.RestoreArgs(restore_type=type(value))
    elif isinstance(value, str):
      return type_handlers.RestoreArgs(restore_type=str)
    else:
      raise ValueError(f'Unsupported type: {type(value)}')

  def _get_sharding_or_layout(value):
    if hasattr(value, 'sharding'):
      if (
          support_layout
          and hasattr(value, 'layout')
          and value.layout.device_local_layout
      ):
        # value is a jax.Array or a jax.ShapeDtypeStruct.
        return value.layout
      else:
        return value.sharding
    else:
      return None

  if sharding_tree is None:
    sharding_tree = jax.tree.map(_get_sharding_or_layout, target)
  if isinstance(target, tree_metadata.TreeMetadata):
    return tree_metadata.build_default_tree_metadata(
        jax.tree.map(_restore_args, target.tree, sharding_tree.tree),
        custom_metadata=target.custom_metadata,
    )
  else:
    return jax.tree.map(_restore_args, target, sharding_tree)
